import re
import math

from requests.api import head
from requests_toolbelt import user_agent
from suplexedlibrary.types.stockx.AlgoliaResponse import StockxAlgoliaResponse
from suplexedlibrary.api.stockx.headers import StockxHeadersUtil
from typing import List
from uuid import uuid4
from pymongo import results
import requests
import random
from collections import OrderedDict
from pprint import pprint
from requests import session
from requests import Response
from datetime import datetime, timedelta
from pytz import timezone
from algoliasearch.search_client import SearchClient
from cloudscraper import create_scraper

from suplexedlibrary.useragents.user_agent_service import get_random_user_agent
from suplexedlibrary.constants.OrderStatusStockx import OrderStatusStockx
from suplexedlibrary.api.graphql import gql_stockx_get_asks
from suplexedlibrary.proxies.provider import get_cloudscraper_datacenter_session
from suplexedlibrary.model.DataSyncAccount import DataSyncAccount
from suplexedlibrary.model.StockxAccount import StockxAccount
from suplexedlibrary.model.AskStockX import AskStockX
from suplexedlibrary.api.tools import header_tools, header_tools_stockx
from suplexedlibrary.database.mongo import nosql
from suplexedlibrary.exceptions import CreateListingException, Exception400, NoLoginExistsException, DeleteListingFailedException, FailedFetchingListings,ListingsChangedWhileFetchingException, Scraping403Exception, DeleteListingNotFoundException, ProductUnavailableForBiddingException, Exception401
from suplexedlibrary.etags import Etags
from suplexedlibrary.database.postgres.postgres import get_session, init_db


PRODUCT_BIDS = 'product_bids'
PRODUCT_ASKS = 'product_asks'
PRODUCT_SALES = 'product_sales'


class StockXApi:
    cache = {
        PRODUCT_SALES: {}
    }
    STOCKX_BROWSE_ENDPOINT = 'https://stockx.com/api/browse'
    STOCKX_HOME_ENDPOINT = 'https://stockx.com'

    def __init__(
        self, 
        account: StockxAccount = None, 
        skip_login: bool = False,
        device_id: str = '17A21C67-DDA0-46AB-AD14-12EF5A069FB9'
    ):
        if account == None:
            print(f'empty construction of {__name__}')
            return

        self.email = account.email
        self.customer_id = account.customer_id
        self.account = account
        self.device_id = device_id

        self.sess = session()
        self.is_logged_in = not skip_login
        self.user_agent = header_tools.get_random_user_agent()
        self.local_cache = {}
        self.local_database_cache = {}
        self.hydrate_db_uuid_cache()
    
        if not skip_login:
            self.set_login_fields()
        
        self.headers_util = StockxHeadersUtil(customer_id=self.customer_id, user_agent=self.user_agent, bearer_token=self.bearer_token)
        self.sess.get('https://stockx.com/', headers={'user-agent': self.user_agent})

    def set_login_fields(self):
            
            db = nosql.get_sneaker_db()
            print(f'\t[+] leveraging login cache for {self.account.email}, {self.account.id}')
            cached_login = db.stockxLoginCache.find_one({
                'email': self.account.email,
                'accountId': self.account.id
            })

            if not cached_login:
                raise NoLoginExistsException

            if 'isMFA' in cached_login and cached_login['isMFA']:
                raise Exception(f'this login has an active mfa session triggered')

            self.bearer_token = cached_login['access_token']
            # self.sess.cookies.update(cached_login['cookies'])
            self.logged_in_headers = header_tools_stockx.get_logged_in_headers(
                self.bearer_token, self.user_agent)
            self.user_agent = cached_login['userAgent']
            self.sess.proxies.update(cached_login['proxy'])
            print(self.sess.proxies)
            db.client.close()

    def print_response_status(self, response: Response):
        url = response.url
        status_code = response.status_code
        print_str = f'{url}: {status_code}'

        if response.status_code != 200:
            print(print_str)
        else:
            print(print_str)

    @staticmethod
    def get_uuid_key(size, stockx_url):
        return f'{stockx_url}//{size}'

    def hydrate_db_uuid_cache(self):

        db = nosql.get_sneaker_db()
        db_uuid_cache_cursor = db.stockx_sku_uuid_store.find({})

        for uuid_object in db_uuid_cache_cursor:
            size = uuid_object['size']
            stockx_url = uuid_object['stx_url']
            key = StockXApi.get_uuid_key(size=size, stockx_url=stockx_url)
            value = uuid_object['uuid']
            
            
            # print(f'adding {key} to cache for {value}')
            self.local_database_cache[key] = value
        db.client.close()

    def get_cookies(self):
        return self.sess.cookies.get_dict()

    def mark_cookies_for_refresh(self):
        db = nosql.get_sneaker_db()
        print(f'\t[!] marking login cache for {self.account.email}, {self.account.id} for refresh')
        db.stockxLoginCache.update_one({
            'email': self.account.email,
            'accountId': self.account.id
        }, {'$set': {
            'refresh': True
        }})

        db.client.close()


    def generate_expiry_stamp(self):
        delta = timedelta(days=30)

        tz = timezone('EST')
        return (datetime.now(tz) +
                delta).isoformat().split('.')[0] + "+0000"

    def get_stockx_sku_uuid(self, stx_url: str, size: str):

        db = nosql.get_sneaker_db()
        sku_scraper_sess = session()
        try:
            cache_key = StockXApi.get_uuid_key(size=size, stockx_url=stx_url)

            if cache_key in self.local_cache:
                print(f'Getting {stx_url} and {size} uuid from local cache')
                db.client.close()
                return self.local_cache[cache_key]

            if cache_key in self.local_database_cache:
                print(f'Getting {stx_url} and {size} uuid from mongo cache')
                db.client.close()
                return self.local_database_cache[cache_key]


            headers = {
                'authority': 'stockx.com',
                'method': 'GET',
                'path': f'/api/products/{stx_url}',
                'scheme': 'https',
                'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
                # 'accept-encoding': 'gzip, deflate, br',
                'accept-language': 'en-US,en;q=0.9',
                'cache-control': 'no-cache',
                'pragma': 'no-cache',
                'sec-ch-ua': '" Not A;Brand";v="99", "Chromium";v="96", "Google Chrome";v="96"',
                'sec-ch-ua-mobile': '?0',
                'sec-ch-ua-platform': '"macOS"',
                'sec-fetch-dest': 'document',
                'sec-fetch-mode': 'navigate',
                'sec-fetch-site': 'none',
                'sec-fetch-user': '?1',
                'upgrade-insecure-requests': '1',
                'user-agent': self.user_agent
            }
            url = f'https://stockx.com/api/products/{stx_url}'
            res = self.sess.get(url, headers=headers)

            if res.status_code != 200:
                print(
                    f'Getting {stx_url} and {size} uuid got {res.status_code}')
                db.client.close()
                return None
            
            json_res = res.json()
            
            if not size and len(json_res['Product']['children']) == 1:
                print('getting uuid for not sized product')
                return list(json_res['Product']['children'].keys())[0]


            for key in json_res['Product']['children']:
                stockx_size = re.sub(
                    '[^0123456789\.]', '', json_res['Product']['children'][key]['shoeSize'])
                my_size = re.sub('[^0123456789\.]', '', size)
                if stockx_size == my_size:
                    
                    db.stockx_sku_uuid_store.update_one(
                        {'size': size, 'stx_url': stx_url}, {
                            '$set': {'size': size, 'stx_url': stx_url, 'uuid': key}},
                        upsert=True
                    )
                    print(f'Getting {stx_url} and {size} uuid found success')
                    db.client.close()
                    return key

            print(
                f'Getting {stx_url} and {size} uuid found nothing in response')
        except Exception as e:
            db.client.close()
            print(e)

    def get_all_buying(self, buy_type: str = 'current', product_type: str = 'sneakers'):
        page = 1
        # limit = 20
        url = 'https://stockx.com/api/customers/' + \
            '{}/buying/'.format(self.customer_id) + buy_type + \
            '?sort=created_at&order=DESC&limit=20&page={}&currency=USD'

        buying_headers = self.headers_util.get_buying_headers()

        res = self.sess.get(url=url.format(page), headers=buying_headers)

        if res.status_code == 401:
            raise Exception401(res.status_code)

        if res.status_code != 200:
            print(f'failed fetching {buy_type} bids for {self.account.email} {res.status_code}')
            return []

        res_json = res.json()
        total = res_json['Pagination']['total']

        total_result_list = res_json['PortfolioItems']
        # result_list = list(filter(
        #     lambda x: x['product']['contentGroup'] == product_type, res_json['PortfolioItems']))
        while len(total_result_list) < total:
            print(f'fetching more portfolio items')
            page += 1
            res = self.sess.get(url=url.format(page), headers=buying_headers)
            res_json = res.json()

            if 'PortfolioItems' not in res_json:
                pprint(res_json)
                continue

            # result_list += list(filter(lambda x: x['product']['contentGroup']
            #                            == product_type, res_json['PortfolioItems']))
            total_result_list += res_json['PortfolioItems']

            if (page * 20) > (total + 20):
                break
        return total_result_list

    def is_historical_cache_up_to_date(self, sales, newest_fetched_sale_date):
        if not newest_fetched_sale_date:
            mongo_db = nosql.get_sneaker_db()
            mongo_db.stockxHistoricalSaleCache.update_one({'user_id': self.account.id}, {'$set': {'newest_fetched_sale_date': datetime.strptime(sales[0]['soldOn'].split('T')[0], '%Y-%m-%d')}}, upsert=True)
            mongo_db.client.close()
            return False
        
        for sale in sales:
            if datetime.strptime(sale['soldOn'].split('T')[0], '%Y-%m-%d') < newest_fetched_sale_date:
                print('no need to proceed further into sales history as we have already found the latest')
                return True
        
        return False

    def get_all_sales(self, sale_type: str = 'current', is_full_scrape: bool = False):
        print(f'full_scrape={is_full_scrape}')
        newest_fetched_sale_date = None
        seen_asks_set = set() # for testing purposes
        if sale_type == 'history' and not is_full_scrape:
            mongo_db = nosql.get_sneaker_db()
            stockx_historical_sale_cache = mongo_db.stockxHistoricalSaleCache.find_one({'user_id': self.account.id})
            mongo_db.client.close()

            if stockx_historical_sale_cache and 'newest_fetched_sale_date' in stockx_historical_sale_cache:
                newest_fetched_sale_date = stockx_historical_sale_cache['newest_fetched_sale_date']

        page = 1
        # limit = 20

        url = f'https://stockx.com/api/customers/{self.customer_id}/selling/{sale_type}?sort=matched_with_date&order=DESC&limit=20&page={page}&currency=USD'
        selling_headers = self.headers_util.get_selling_headers(request_url=url)
        pprint(selling_headers)
        res = self.sess.get(url=url, headers=selling_headers)

        if res.status_code == 302:
            print('received 302, content not modified... returning none')
            return None

        self.headers_util.write_etag(url, res.headers)

        if res.status_code == 304:
            print(f'users {sale_type}-sales not modified, we can skip')
            return []

        if res.status_code == 401:
            raise Exception401(res.status_code)

        if res.status_code != 200:
            print(f'failed fetching {sale_type} sales for {self.account.email} {res.status_code}')
            raise FailedFetchingListings(f'sale_type:{sale_type} status_code:{res.status_code}')

        res_json = res.json()
        if 'Pagination' not in res_json:
            print('No asks found on this account')
            return []

        total = res_json['Pagination']['total']
        print(f'fetching a total of {total} listings')
        result_list = res_json['PortfolioItems']
        for item in res_json['PortfolioItems']:
            chain_id = item['chainId']
            if chain_id in seen_asks_set:
                print(f'found a duplicate chain_id={chain_id} on page={page}')
            else:
                seen_asks_set.add(chain_id)


        if sale_type == 'history' and not is_full_scrape:
            is_up_to_date = self.is_historical_cache_up_to_date(result_list, newest_fetched_sale_date)
            if is_up_to_date:
                return result_list
        else:
            is_up_to_date = False

        while len(result_list) < total and not is_up_to_date:
            page += 1
            url = f'https://stockx.com/api/customers/{self.customer_id}/selling/{sale_type}?sort=matched_with_date&order=DESC&limit=20&page={page}&currency=USD'
            res = self.sess.get(url=url, headers=selling_headers)

            if res.status_code != 200:
                print(f'failed fetching {sale_type} sales for {self.account.email} {res.status_code}')
                raise FailedFetchingListings(f'sale_type:{sale_type} status_code:{res.status_code}')

            res_json = res.json()
            for item in res_json['PortfolioItems']:
                chain_id = item['chainId']
                if chain_id in seen_asks_set:
                    print(f'found a duplicate chain_id={chain_id} on page={page}')
                else:
                    seen_asks_set.add(chain_id)

            result_list += res_json['PortfolioItems']
            
            # total = res_json['Pagination']['total']
            new_total = res_json['Pagination']['total']
            if total != new_total:
                print(f'User is making edits off of sync, stop getting data, stockx_email:{self.account.email}')
                raise FailedFetchingListings(f'sale_type:{sale_type} total mismatch, {total} != {new_total}')
            '''
            if total != new_total:
                if new_total == 0:
                    print(f'we reached 0 which is the end')
                    break

                print(f'total:{total} != new_total:{new_total} -- FAILING -> THE LIST CHANGED')
                raise ListingsChangedWhileFetchingException
                # return self.get_all_sales(sale_type=sale_type) 
            '''
            if sale_type == 'history' and not is_full_scrape:
                is_up_to_date = self.is_historical_cache_up_to_date(result_list, newest_fetched_sale_date)

                if is_up_to_date:
                    mongo_db = nosql.get_sneaker_db()
                    mongo_db.stockxHistoricalSaleCache.update_one({'user_id': self.account.id}, {'$set': {'newest_fetched_sale_date': datetime.strptime(result_list[0]['matchedWithDate'].split('T')[0], '%Y-%m-%d')}}, upsert=True)
                    mongo_db.client.close()
                    return result_list

            print(f'{len(result_list)}/{total} listings fetched so far on page {page}')

        if sale_type == 'history' and not is_full_scrape and len(result_list) > 0:
            mongo_db = nosql.get_sneaker_db()
            mongo_db.stockxHistoricalSaleCache.update_one({'user_id': self.account.id}, {'$set': {'newest_fetched_sale_date': datetime.strptime(result_list[0]['matchedWithDate'].split('T')[0], '%Y-%m-%d')}}, upsert=True)
            mongo_db.client.close()

        filtered_results = []
        for item in result_list:
            if item['chainId'] in seen_asks_set:
                seen_asks_set.remove(item['chainId'])
                filtered_results.append(item)

        return result_list

    def delete_all_asks(self, sell_type: str = 'CURRENT'):
        gql_endpoint = 'https://stockx.com/p/e'
        newest_fetched_sale_date = None
        after = None
        body = {
            "operationName": "FetchCustomerAsks",
            "variables": {
                "after"   : after,
                "limit"   : 20,
                "state"   : sell_type,
                "sort"    : "LISTED_AT",
                "order"   : "DESC",
                "query"   : "",
                "currency": "USD",
                "country" : "US"
            },
            "query": gql_stockx_get_asks.query
        }
        headers = self.headers_util.get_delete_headers()
        gql_res = self.sess.post(gql_endpoint, headers=headers, json=body)

        if gql_res.status_code == 403:
            raise Scraping403Exception()
        if gql_res.status_code != 200:
            raise FailedFetchingListings(f'Failed fetching listings status_code={gql_res.status_code}')

        response_data = gql_res.json()['data']

        asks = response_data['viewer']['asks']['edges']
        page_info = response_data['viewer']['asks']['pageInfo']
        after = page_info['endCursor'] if 'endCursor' in page_info else None

        result = []
        for ask in asks:
            ask = ask['node']
            state = ask["state"]
            result.append({
                **ask, 'status': vars(OrderStatusStockx).get(f'_{state}'), 'chainId': ask['id']
            })
            print(f'{ask["productVariant"]["product"]["urlKey"]}={state}')
            self.remove_ask(ask['id'])


        while page_info['hasNextPage'] is True:
            body = {
                "operationName": "FetchCustomerAsks",
                "variables": {
                    "after"   : after,
                    "limit"   : 100,
                    "state"   : sell_type,
                    "sort"    : "LISTED_AT",
                    "order"   : "DESC",
                    "query"   : "",
                    "currency": "USD",
                    "country" : "US"
                },
                "query": gql_stockx_get_asks.query
            }

            gql_res = self.sess.post(gql_endpoint, headers=headers, json=body)

            if gql_res.status_code == 403:
                raise Scraping403Exception()
            if gql_res.status_code != 200:
                raise FailedFetchingListings(f'Failed fetching listings status_code={gql_res.status_code}')

            response_data = gql_res.json()['data']

            asks = response_data['viewer']['asks']['edges']
            page_info = response_data['viewer']['asks']['pageInfo']
            after = page_info['endCursor'] if 'endCursor' in page_info else None
            for ask in asks:
                ask = ask['node']
                state = ask["state"]
                result.append({
                    **ask, 'status': vars(OrderStatusStockx).get(f'_{state}'), 'chainId': ask['id']
                })
                
                print(f'{ask["productVariant"]["product"]["urlKey"]}={state}')
                self.remove_ask(ask['id'])



        print(f'found {len(result)} asks')
        return result


    def get_sales_history_gql(self, sell_type: str = 'HISTORICAL', is_full_scrape: bool = True):
        selling_endpoint = 'https://stockx.com/p/e'
        newest_fetched_sale_date = None
        seen_asks_set = set() # for testing purposes
        if not is_full_scrape and sell_type == 'HISTORICAL':
            mongo_db = nosql.get_sneaker_db()
            stockx_historical_sale_cache = mongo_db.stockxHistoricalSaleCache.find_one({'user_id': self.account.id})
            mongo_db.client.close()

            if stockx_historical_sale_cache and 'newest_fetched_sale_date' in stockx_historical_sale_cache:
                newest_fetched_sale_date = stockx_historical_sale_cache['newest_fetched_sale_date']


        after = None
        body = {
            "operationName": "FetchCustomerAsks",
            "variables": {
                "after"   : after,
                "limit"   : 100,
                "state"   : sell_type,
                "sort"    : "LISTED_AT",
                "order"   : "DESC",
                "query"   : "",
                "currency": "USD",
                "country" : "US"
            },
            "query": gql_stockx_get_asks.query
        }
        headers = self.headers_util.get_selling_headers_gql(request_url=f'https://stockx.com/api/graphql:{sell_type}')
        gql_res = self.sess.post(selling_endpoint, headers=headers, json=body)
        # self.headers_util.write_etag(request_url=f'https://stockx.com/api/graphql:{sell_type}', headers=gql_res.headers)


        if gql_res.status_code == 401:
            raise Exception401(gql_res.status_code)

        if gql_res.status_code != 200:
            raise FailedFetchingListings(f'Failed fetching listings status_code={gql_res.status_code}')

        response_data = gql_res.json()['data']

        asks = response_data['viewer']['asks']['edges']
        page_info = response_data['viewer']['asks']['pageInfo']
        after = page_info['endCursor'] if 'endCursor' in page_info else None

        result = []
        for ask in asks:
            try:
                ask = ask['node']
                state = ask["state"]
                result.append({
                    **ask, 'status': vars(OrderStatusStockx).get(f'_{state}'), 'chainId': ask['id']
                })
                print(f'{ask["productVariant"]["product"]["urlKey"]}={state}')
            except Exception as e:
                print(e)

        while page_info['hasNextPage'] is True:
            body = {
                "operationName": "FetchCustomerAsks",
                "variables": {
                    "after"   : after,
                    "limit"   : 100,
                    "state"   : sell_type,
                    "sort"    : "LISTED_AT",
                    "order"   : "DESC",
                    "query"   : "",
                    "currency": "USD",
                    "country" : "US"
                },
                "query": gql_stockx_get_asks.query
            }

            gql_res = self.sess.post(selling_endpoint, headers=headers, json=body)

            if gql_res.status_code != 200:
                raise FailedFetchingListings(f'Failed fetching listings status_code={gql_res.status_code}')

            response_data = gql_res.json()['data']

            asks = response_data['viewer']['asks']['edges']
            page_info = response_data['viewer']['asks']['pageInfo']
            after = page_info['endCursor'] if 'endCursor' in page_info else None
            for ask in asks:
                try:
                    ask = ask['node']
                    state = ask["state"]
                    result.append({
                        **ask, 'status': vars(OrderStatusStockx).get(f'_{state}'), 'chainId': ask['id']
                    })
                    print(f'{ask["productVariant"]["product"]["urlKey"]}={state}')
                except Exception as e:
                    print(e)

            if sell_type == 'HISTORICAL' and not is_full_scrape:
                is_up_to_date = self.is_historical_cache_up_to_date(result, newest_fetched_sale_date)
                if is_up_to_date:
                    return result                    

        print(f'found {len(result)} asks')
        return result

    def remove_ask(self, chain_id: str) -> str:
        return self.remove_stockx_order(chain_id)

    def remove_stockx_order(self, chain_id: str):

        payload = {
            "chain_id": chain_id,
            "notes": "Customer Deleted Bid"
        }

        headers = self.headers_util.get_remove_order_headers(chain_id=chain_id)

        if chain_id == None:
            print(f'Order has no chain_id')
            return

        # sess = get_proxied_session()
        res = self.sess.delete(url=f'https://stockx.com/api/portfolio/{chain_id}', headers=headers, json=payload)

        if res.status_code == 200:
            print(f'Successfully removed ask for {chain_id}')
            return 200
        elif res.status_code == 400:
            print(f'this ask is already removed {chain_id}')
            return 400
        elif res.status_code == 404:
            raise DeleteListingNotFoundException(res.status_code)
        elif res.status_code == 403:
            raise Scraping403Exception(res.status_code)
        else:
            print(f'{res.status_code} received when deleting ask')
            print(res.content)
            raise DeleteListingFailedException(res.status_code)
        
    def update_ask(
        self,
        ask: AskStockX,
        amount: int,
        sell_floor: int,
    ):
        if sell_floor != None:
            if amount < sell_floor:
                print("We tried to go below the sell floor for: " + str(ask.exchange_ask_id) +
                            ', new_ask = ' + str(amount) + ' , sell_floor = ' + str(sell_floor))
                return

        url = 'https://stockx.com/api/portfolio?a=ask'
        payload = {
            "PortfolioItem": {
                "localAmount": amount,
                "expiresAt": self.generate_expiry_stamp(),
                "skuUuid": ask.stx_sku_uuid,
                "localCurrency": "USD",
                "meta": {"discountCode": ""},
                "chainId": ask.exchange_ask_id
            }}

        update_response = self.sess.post(
            url, json=payload, headers=self.logged_in_headers)

        print(f'Success updating ask {ask.exchange_ask_id}')

    def update_ask_raw(
        self,
        exchange_ask_id: str,
        stx_sku_uuid: str,
        amount: int,
        portfolio_item: dict,
        sell_floor: int = None,
    ):
        url = 'https://stockx.com/api/portfolio?a=ask'
        payload = {
            "action": "ask",            
            "PortfolioItem": {
                "localAmount": amount,
                "expiresAt": self.generate_expiry_stamp(),
                "skuUuid": stx_sku_uuid,
                "localCurrency": "USD",
                "meta": {"discountCode": ""},
                "chainId": exchange_ask_id
            },
            'item': portfolio_item
        }

        update_response = self.sess.post(
            url, json=payload, headers={
                'authority': 'stockx.com',
                'method': 'POST',
                'path': '/api/portfolio?a=ask',
                'scheme': 'https',
                'accept': '*/*',
                # 'accept-encoding': 'gzip, deflate, br',
                'accept-language': 'en-US,en;q=0.9',
                'appos': 'web',
                'appversion': '0.1',
                'authorization': f'Bearer {self.bearer_token}',
                'content-type': 'application/json',
                'origin': 'https://stockx.com',
                'referer': f'https://stockx.com/sell/{exchange_ask_id}',
                'sec-ch-ua': '" Not A;Brand";v="99", "Chromium";v="96", "Google Chrome";v="96"',
                'sec-ch-ua-mobile': '?0',
                'sec-ch-ua-platform': '"macOS"',
                'sec-fetch-dest': 'empty',
                'sec-fetch-mode': 'cors',
                'sec-fetch-site': 'same-origin',
                'user-agent': self.user_agent,
                'x-fraud-cookie-id': '798bfaa072db4fbba1c1bc75cb335198_1643076428758__UDF43_13ck_tt',
                'x-requested-with': 'XMLHttpRequest',
                'x-session-id': '36c94ada-949b-44e5-8593-8b36227a98d9',
            })
        
        if update_response.status_code != 200:

            if update_response.status_code == 403:
                raise Scraping403Exception(update_response.status_code)

            if update_response.status_code == 400:
                raise Exception400()
            
            raise Exception(update_response.status_code)

        print(f'Success updating ask {exchange_ask_id}')
        return update_response.status_code


    def update_bulk_asks(
        self,
        items: List[dict]
    ):
        portfolio_items = []
        for item in items:
            portfolio_items.append({'PortfolioItem': {
                "skuUuid": item['stx_sku_uuid'],
                'chainId': item['chain_id'],
                "localAmount": item['new_ask'],

                "expiresAt": self.generate_expiry_stamp(),
                "localCurrency": "USD",
                "meta": {"discountCode": ""},
                'quantity': 1
            }})
        headers = self.headers_util.get_update_bulk_ask_headers()
        url = 'https://stockx.com/api/portfolio/multi?a=ask'

        response = self.sess.post(url, headers=headers, json=portfolio_items)

        if response.status_code != 200:
            print(f'{response.status_code} received while placing bulk asks')
            print(response.content)
            print(response.json())
            err_res_json = response.json()
            if err_res_json['message'] == 'This product is unavailable to sell':
                raise ProductUnavailableForBiddingException(f'{response.status_code} received while placing ask for bulk')

            if response.status_code == 403:
                raise Scraping403Exception(response.status_code)
                
            raise CreateListingException(response.status_code)
        
        return response.status_code
    
    def place_new_bulk_asks(
        self,
        items: List[dict]
    ) -> int:
        headers = self.headers_util.get_update_bulk_ask_headers()
        url = 'Request URL: https://stockx.com/api/portfolio/multi?a=ask'
        portfolio_items = []
        expiry = self.generate_expiry_stamp()
        for item in items:
            portfolio_items.append({'PortfolioItem': {
                "localAmount": item['new_ask'],
                "expiresAt": self.generate_expiry_stamp(),
                "skuUuid": item['stx_sku_uuid'],
                "localCurrency": "USD",
                "meta": {"discountCode": ""}
            }})

        response = self.sess.post(url, headers=headers, json=portfolio_items)

        if response.status_code != 200:
            print(f'{response.status_code} received while placing bulk asks')
            print(response.content)
            print(response.json())
            err_res_json = response.json()
            if err_res_json['message'] == 'This product is unavailable to sell':
                raise ProductUnavailableForBiddingException(f'{response.status_code} received while placing ask for bulk')

            if response.status_code == 403:
                raise Scraping403Exception(response.status_code)
                
            raise CreateListingException(response.status_code)

    def place_new_ask(
        self,
        style_id: str,
        stx_url: str,
        size: str,
        new_ask: int,
        sell_floor: int = None,
        uuid=None
    ) -> int:
        if sell_floor != None:
            if new_ask < sell_floor:
                print("We tried to go below the sell floor for while placing a new ask: " + str(style_id) + ', new_ask = ' + str(new_ask) + ' , sell_floor = ' + str(sell_floor))
                return

        if uuid == None:
            stx_sku_uuid = self.get_stockx_sku_uuid(stx_url=stx_url, size=str(size.replace('.0', '')))
            if stx_sku_uuid == None:
                print('\t[!] No size in children for ' + style_id)
                raise Exception(f'Could not fetch uuid for {stx_url}')
        else:
            stx_sku_uuid = uuid

        url = 'https://stockx.com/api/portfolio?a=ask'
        payload = {
            "PortfolioItem": {
                "localAmount": new_ask,
                "expiresAt": self.generate_expiry_stamp(),
                "skuUuid": stx_sku_uuid,
                "localCurrency": "USD",
                "meta": {"discountCode": ""}
            }}
        response = self.sess.post(
            url, json=payload, headers={
                'authority': 'stockx.com',
                'method': 'POST',
                'path': '/api/portfolio?a=ask',
                'scheme': 'https',
                'accept': '*/*',
                # 'accept-encoding': 'gzip, deflate, br',
                'accept-language': 'en-US,en;q=0.9',
                'appos': 'web',
                'appversion': '0.1',
                'authorization': f'Bearer {self.bearer_token}',
                'content-length': '190',
                'content-type': 'application/json'                ,
                'origin': 'https://stockx.com',
                'referer': f'https://stockx.com/sell/{stx_url}?size={size}',
                'sec-ch-ua': '" Not A;Brand";v="99", "Chromium";v="96", "Google Chrome";v="96"',
                'sec-ch-ua-mobile': '?0',
                'sec-ch-ua-platform': '"macOS"',
                'sec-fetch-dest': 'empty',
                'sec-fetch-mode': 'cors',
                'sec-fetch-site': 'same-origin',
                'user-agent': self.user_agent,
                'x-fraud-cookie-id': '798bfaa072db4fbba1c1bc75cb335198_1643022397384__UDF43_13ck_tt',
                'x-requested-with': 'XMLHttpRequest',
                'x-session-id': 'f30676fd-c232-4816-b205-bedc4a872916',
            })

        if response.status_code != 200:
            print(f'{response.status_code} received while placing ask for {stx_url}')
    
            if response.status_code == 403:
                raise Scraping403Exception(response.status_code)
    
            err_res_json = response.json()
            if response.status_code == 400 or err_res_json['message'] == 'This product is unavailable to sell':
                raise ProductUnavailableForBiddingException(f'{response.status_code} received while placing ask for {stx_url}')

                
            raise CreateListingException(response.status_code)

        print(f'Success creating new ask for {stx_url}')
        json_res = response.json()
        chain_id = json_res['PortfolioItem']['chainId']
        
        return {
            'chain_id': chain_id,
            'stx_sku_uuid': stx_sku_uuid,
            'portfolio_item': json_res['PortfolioItem']
        }


    def place_new_bid(self, stx_sku_uuid: str, new_bid: int):

        print(f'Placing new bid for ({stx_sku_uuid}) at bid of {new_bid}')

        url = "https://stockx.com/api/portfolio?a=bid"
        payload = {
            "PortfolioItem": {
                "localAmount": int(new_bid),
                "expiresAt": self.generate_expiry_stamp(),
                "skuUuid": stx_sku_uuid,
                "localCurrency": "USD",
                "meta": {
                    "discountCode": ""
                }
            },
            "action":"bid",
        }
        temp_headers = self.logged_in_headers
        res = self.sess.post(
            url, json=payload, headers=self.logged_in_headers)
        
        if res.status_code != 200:
            # if res.status_code == 400:
            pprint(res.json())
            print(f'{res.status_code}: Failed placing bid for {stx_sku_uuid}')
            return

        json_res = res.json()
        print(f'Success creating new bid of {json_res["PortfolioItem"]["chainId"]}')
        return json_res["PortfolioItem"]["chainId"]


    def update_bid(self, chain_id: str, stx_sku_uuid: str, new_bid: int):
        try:
            url = 'https://stockx.com/api/portfolio?a=bid'
            payload = {
                "PortfolioItem": {
                    "localAmount": new_bid,
                    "expiresAt": self.generate_expiry_stamp(),
                    "skuUuid": stx_sku_uuid,
                    "localCurrency": "USD",
                    "meta": {"discountCode": ""},
                    "chainId": chain_id
                },
            }
            # sess = get_proxied_session()
            response = self.sess.post(
                url, json=payload, headers=self.logged_in_headers)

            
            if response.status_code != 200:
                print(f'{response.status_code}: Failed placing bid for {stx_sku_uuid}')
            
            return response.status_code
            
        except Exception as e:
            print(f'{e}: occurred updating bid for {stx_sku_uuid}')
            raise e


    def get_customer_id(self) -> str:
        print(f'fetching customer id for {self.bearer_token}')
        url = 'https://stockx.com/api/users/me/selling?currency=USD'
        headers = self.headers_util.get_customer_id_headers()



        res = self.sess.get(url=url, headers=headers)
        if res.status_code != 200:
            print(res.content)
            return None
        res_json = res.json()
        account_id = res_json['data'][0]['id']
        print(f'found {account_id} as account id')

        return account_id

    def get_order_details(self, chain_id: str):
        try:
            url = f'https://stockx.com/api/portfolioItems/{chain_id}?includes=pricing,payment,payout,shipping,product,meta&country=US'            
            
            headers = self.headers_util.get_order_details_headers(request_url=url)


            order_details_res = self.sess.get(url, headers=headers)

            if order_details_res.status_code != 200:
                print(order_details_res.content)
                print(order_details_res.status_code)
            
            if order_details_res.status_code == 302:
                print('caught 302 with etags')
                return None

            self.headers_util.write_etag(url, order_details_res.headers)

            return order_details_res.json()

        except Exception as e:
            print(e)
            return None

    def get_seller_fees(self) -> tuple:
        try:
            url = 'https://stockx.com/api/users/me/rewards'
            headers = self.headers_util.get_seller_fees_headers()
            me_response = self.sess.get(url, headers=headers)

            me_json = me_response.json()
            seller_fee = me_json['data'][0]['attributes']['level']['current']['fee']
            seller_level = me_json['data'][0]['attributes']['level']['current']['level']
            return (seller_fee, seller_level)

        except Exception as e:
            print(e)

    def test_api_login(self):
        url = 'https://stockx.com/api/users/me/rewards'
        headers = self.logged_in_headers
        me_response = self.sess.get(url, headers=headers)

        if me_response.status_code != 200:
            raise Exception(f'Failed test')

    def get_chain_id_details(self, chain_id: str):
        url = f'https://stockx.com/api/portfolioItems/{chain_id}?country=US'
        selling_headers = self.headers_util.get_chain_id_headers(request_url=url)
        res = self.sess.get(url, headers=selling_headers)
        if res.status_code != 200:
            raise Exception(f'[{res.status_code}]: Failed getting chain id details')
        
        self.headers_util.write_etag(request_url=url, headers=res.headers)
        listing_details = res.json()
        return listing_details['PortfolioItem']

    def get_stockx_sales(uuid: str):
        end_date = '2021-01-07'
        limit = 500
        page = 1
        total_pages = None

        sales = []
        while True:
            url = f'https://stockx.com/api/products/{uuid}/activity?state=480&currency=USD&limit={limit}&page={page}&sort=createdAt&order=DESC&country=US'
            print(f'\tfetched {page}/{total_pages}')
            print(f'fetching {url}')
            res = requests.get(url, headers=header_tools_stockx.light_headers)

            if res.status_code != 200:
                print(f'failed getting historical stales with status={res.status_code}')
    
            data = res.json()
            if 'ProductActivity' not in data:
                break

            sales += data['ProductActivity']
            total_pages = data['Pagination']['total']
            page += 1

        print(f'found {len(sales)} sales for {uuid}')
        return sales

    def price_product_by_id(self, product_id: str) -> float:
        url = f'https://stockx.com/api/products/{product_id}/market?currency=USD&country=US'
        headers = self.headers_util.get_pricing_headers()
        res = self.sess.get(url, headers=headers)

        if res.status_code != 200:
            print(res.status_code)
            print(res.content)
            raise Exception('Failed getting pricing data')

        
        res_json = res.json()
        market_data = res_json['Market']

        return {
            'lowestPrice': market_data['lowestAsk'],
            'highestOffer': market_data['highestBid']
        }

    @staticmethod
    def get_stockx_url_from_algolia_search(query: str):
        client = SearchClient.create('XW7SBCT9V6', '6b5e76b49705eb9f51a06d3c82f7acee')
        index = client.init_index('products')
        objects = index.search(query)
        stockx_url = objects.get('hits')[0]['url']
        client.close()
        return stockx_url

    @staticmethod
    def get_stockx_pricing(query: str):
        sess = session()
        limit = 5
        stockx_url = StockXApi.get_stockx_url_from_algolia_search(query)

        cache_key = f'stockx_asks:{stockx_url}'
        if cache_key in StockXApi.cache:
            return StockXApi.cache[cache_key]
        
        try:
            url = f'https://stockx.com/api/products/{stockx_url}/activity?state=400&currency=USD&limit={limit}&page=1&sort=amount&order=ASC&country=US'
            res = sess.get(url, headers=header_tools_stockx.light_headers, timeout=10)

            if res.status_code == 200:
                # print('\t[$] Success fetching ask data')
                result = res.json()

                if 'ProductActivity' not in result:
                    # self.cache[PRODUCT_ASKS] = []
                    result = []
                else:
                    result = result['ProductActivity']

                StockXApi.cache[cache_key] = result

                return result
            else:
                print(f'{res.status_code} Received while fetching stockx ask stack')

        except TimeoutError as e:
            print(f'{e}: occurred while fetching stockx ask stack')
            StockXApi.cache[cache_key] = []

        except Exception as e:
            print(f'{e}: occurred while fetching stockx ask stack')

        return []

    @staticmethod
    def get_stockx_chart_sales(uuid: str, start_date: str = 'all', end_date: datetime = datetime.now()):
        end_date_str = end_date.strftime('%Y-%m-%d')
        start_date_str = start_date.strftime('%Y-%m-%d')
        url = f'https://stockx.com/api/products/{uuid}/chart?start_date={start_date_str}&end_date={end_date_str}&intervals=10000&currency=USD&country=US'
        res = requests.get(url, headers=header_tools_stockx.light_headers)

        if res.status_code != 200:
            print(f'failed getting chart sales for {uuid}')
            return {}

        return res.json()

    @staticmethod
    def get_stockx_chart_sales_parsed(uuid: str, size: str, style_id: str, start_date: str = 'all', end_date: datetime = datetime.now()):
        cache_key = f'parsed-chart:{uuid}:{size}'
        if cache_key in StockXApi.cache:
            return StockXApi[cache_key]

        data = StockXApi.get_stockx_chart_sales(uuid=uuid, start_date=start_date, end_date=end_date)

        y_axis = data['series'][0]['data']
        x_axis = data['xAxis']['categories']

        data_points = []
        for i, date_range_str in enumerate(x_axis):
            data_points.append({
                'data_range': date_range_str,
                'average_sale_price': y_axis[i],
                'uuid': uuid,
                'style_id': style_id,
                'size': size
            })

        StockXApi.cache[cache_key] = data_points
        return data_points

    @staticmethod
    def get_all_product_children(stockx_url: str):
        headers = {
            "user-agent": get_random_user_agent()
        }

        headers = header_tools_stockx.light_headers
        url = f'https://stockx.com/api/products/{stockx_url}'
        res = get_cloudscraper_datacenter_session().get(url, headers=headers)

        if res.status_code != 200:
            print(f'\t\tFailed fetching {stockx_url} received status_code={res.status_code}')
            return {}
        json_res = res.json()

        product_children = json_res['Product']['children']
        return product_children

    @staticmethod
    def get_stockx_sku_uuid_static(stx_url: str, size: str):
        cache_key = f'stockxUUID:{stx_url}-{size}'

        if cache_key in StockXApi.cache:
            return StockXApi.cache[cache_key]
    
        db = nosql.get_sneaker_db()
        uuid_object = db.stockxUUIDCache.find_one({'uuid_key': stx_url})
        if uuid_object:
            return uuid_object['uuid']

        headers = {
            "user-agent": get_random_user_agent()
        }

        url = f'https://stockx.com/api/products/{stx_url}'
        res = get_cloudscraper_datacenter_session().get(url, headers=headers)

        if res.status_code == 403:
            raise Scraping403Exception(f'Failed scraping status_code={res.status_code}')
        elif res.status_code != 200:
            raise Exception('Bad request made')

        json_res = res.json()

        for key in json_res['Product']['children']:
            stockx_size = re.sub(
                '[^0123456789\.]', '', json_res['Product']['children'][key]['shoeSize'])
            my_size = re.sub('[^0123456789\.]', '', size)
            if stockx_size == my_size:
                
                db.stockx_sku_uuid_store.update_one(
                    {'size': size, 'stx_url': stx_url}, {
                        '$set': {'size': size, 'stx_url': stx_url, 'uuid': key}},
                    upsert=True
                )
                print(f'Getting {stx_url} and {size} uuid found success')
                db.client.close()
                StockXApi.cache[cache_key] = key
                db.stockxUUIDCache.insert_one({'uuid_key': stx_url, 'uuid': key})
                return key

        print(
            f'Getting {stx_url} and {size} uuid found nothing in response')

    @staticmethod
    def search_stockx_algolia(search_term: str):
        client = SearchClient.create('XW7SBCT9V6', '6b5e76b49705eb9f51a06d3c82f7acee')
        index = client.init_index('products')
        objects = index.search(search_term)
        client.close()

        return objects.get('hits')      

    @staticmethod
    def get_stockx_product_data(url_key: str):
        headers = {
            'Content-Type': 'application/json',
            'App-Version': '3.7.3.12378',
            'x-api-key': '99WtRZK6pS1Fqt8hXBfWq8BYQjErmwipa3a0hYxX',
            'App-Name': 'StockX-iOS',
            'JWT-Authorization': '',
            'Accept-Language': 'en-us',
            'User-Agent': 'StockX/12378 CFNetwork/889.9 Darwin/17.2.0',        
        }

        url = f'https://stockx.com/api/products/{url_key}'
        res = get_cloudscraper_datacenter_session().get(url, headers=headers)

        return res.json()

    @staticmethod
    def get_stockx_product_sizes(url_key: str):
        data = StockXApi.get_stockx_product_data(url_key)

        size_dict = data['Product']['children']
        result = []
        for key in size_dict.keys():
            result.append({
                'size': size_dict[key]['shoeSize'],
                'stockx_uuid': key
            })
        return result


    @staticmethod
    def get_typed_algolia_search_response(query: str) -> StockxAlgoliaResponse:
        client = SearchClient.create('XW7SBCT9V6', '6b5e76b49705eb9f51a06d3c82f7acee')
        index = client.init_index('products')
        response = index.search(query)
        
        client.close()

        return StockxAlgoliaResponse(**response)
        

    @staticmethod
    def get_browse(
        _tags: str = None,
        browse_index: str = None,
        gender: str = 'men',
        order: str = 'DESC',
        range_upper: int = None,
        range_lower: int = None,
    ):
        local_sess = create_scraper()
        local_sess.proxies.update({
            'http': f'http://user-sp56852893:Milo2021!@gate.dc.smartproxy.com:{random.randint(20000, 20999)}'

        })
        params = [
            ('browseVerticals', 'sneakers'),
            ('gender', gender),
            ('order', order)
        ]
        if range_upper and range_lower:
            params.append(('market.lowestAsk', f'range({range_lower}|{range_upper})'))
        if browse_index:
            params.append(('sort', browse_index))
        if _tags:
            params.append(('_tags', _tags))

        res = local_sess.get(url=StockXApi.STOCKX_BROWSE_ENDPOINT, params=params)
        if res.status_code != 200:
            if res.status_code == 403:
                raise Scraping403Exception()
            if res.status_code == 400:
                raise Exception400()

        return res.json()


    def get_chain_id(self, url: str):
        res = self.sess.get(url, headers={
                'authority': 'stockx.com',
                'method': 'POST',
                'scheme': 'https',
                'accept': '*/*',
                'accept-language': 'en-US,en;q=0.9',
                'appos': 'web',
                'appversion': '0.1',
                'authorization': f'Bearer {self.bearer_token}',
                'content-type': 'application/json',
                'origin': 'https://stockx.com',
                'sec-fetch-dest': 'empty',
                'sec-fetch-mode': 'cors',
                'sec-fetch-site': 'same-origin',
                'user-agent': self.user_agent,
                'x-requested-with': 'XMLHttpRequest',
                'x-session-id': '681a08d7-ef89-434a-999b-7d01a05267ca',
        })
        if res.status_code != 200:
            print(f'received status_code={res.status_code}')
            print(url)
            raise Exception(res.status_code)

        return res

    def get_user_notifications_gql(self):
        gql_endpoint = 'https://stockx.com/api/graphql'
        query = {
            "variables": {
                "first": 299,
                "deviceId": self.device_id
            },
            "query": "query FetchInboxMessages($deviceId: String!, $first: Int!, $after: String, $sort: InboxSortInput) {\n  viewer {\n    __typename\n    inbox(deviceId: $deviceId, first: $first, after: $after, sort: $sort) {\n      __typename\n      edges {\n        __typename\n        node {\n          __typename\n          id\n          deliveredAt\n          expiresAt\n          title\n          subtitle\n          isRead\n          imageUrl\n          link\n        }\n      }\n      pageInfo {\n        __typename\n        totalCount\n        hasNextPage\n        hasPreviousPage\n        endCursor\n        startCursor\n      }\n    }\n  }\n}"
        }
        url = gql_endpoint
        headers = self.headers_util.get_notifications_headers_gql(self.device_id)
        res = self.sess.post(url=url, json=query, headers=headers)

        if res.status_code == 403:
            raise Scraping403Exception()
        if res.status_code != 200:
            raise Exception(f'Failed fetching notifications status_code={res.status_code}')

        notifications = res.json()['data']['viewer']['inbox']['edges']
        notifications_list = []
        for edge in notifications:
            if  'new low ask' in edge['node']['title'].lower():
                notifications_list.append({
                    'id': edge['node']['id'],
                    'delivered_at': edge['node']['deliveredAt'],
                    'chain_id': edge['node']['link'].split('https://stockx.com/selling/')[1].split('?')[0],
                    'new_lowest_ask_amount': int(edge['node']['subtitle'].split('The new lowest Ask is $')[1].split('for')[0].replace(',', ''))
                })


        pprint(notifications_list)
        print(f'found {len(notifications_list)}')
        return notifications_list

if __name__ == '__main__':




    res = StockXApi.get_typed_algolia_search_response('reebok')
    from suplexedlibrary.database.postgres.postgres import get_new_session
    init_db()
    stockx_api = StockXApi(get_new_session().query(StockxAccount).filter(StockxAccount.email == 'sneakervaultecho@gmail.com').first(), skip_login=False)
    
    stockx_api.get_sales_history_gql(sell_type='CURRENT')
    stockx_api.get_user_notifications_gql()