import sys
import time
import os
import pickle
from datetime import datetime
from pprint import pprint
import requests


from suplexedlibrary.proxies.provider import get_whitelisted_stockx_proxy
from suplexedlibrary.database.mongo import nosql
from suplexedlibrary.util import shoe_sizes
from suplexedlibrary.types.SuplexedPricingResponse import SuplexedPricingResponse, SuplexedPricingResponse_Meta, SuplexedPricingResponse_Marketplaces, SuplexedPricingResponse_Listings_Summary, SuplexedPricingResponse_Pricing_Summary_Details

class SuplexedSizeResponse:
    def __init__(
        self,
        meta,
        marketplaces,
        pricing,
    ) -> None:
        self.meta = meta
        self.marketplaces = marketplaces
        self.pricing = pricing
        
    meta: SuplexedPricingResponse_Meta
    marketplaces: SuplexedPricingResponse_Marketplaces
    pricing: SuplexedPricingResponse_Listings_Summary[int]

class SuplexApi:
    query_overrides = {
        'DA2728-100'
    }

    def __init__(self):
        self.local_cache = {}
        self.local_cache_v2 = {}
        self.sess = requests.session()


    def get_goat_product_url(self, style_id: str) -> str:
        cache_key = f'product_url:{style_id}'
        
        if cache_key in self.local_cache:
            print(f'using cached url for {cache_key}')
            return self.local_cache[cache_key]

        suplexed_url = os.environ.get('SUPLEXED_PRICING_API_URL') or 'https://suplexed-pricing-api.suplexed.io'

        url = f'{suplexed_url}/{style_id}/prices'

        if cache_key in self.local_cache_v2:
            return self.local_cache_v2[cache_key]

        res = self.sess.get(url)
        
        if res.status_code != 200:
            raise Exception(f'[status_code={res.status_code}] Failed Fetching Suplexed Data for style_id={style_id}')

        res_json = res.json()
        result = res_json['marketplaces']['goat']['url'].split('https://www.goat.com/sneakers/')[1]
        # self.local_cache_v2[cache_key] = result

        return result


    def get_pricing_data_v2(self, style_id: str):
        suplexed_url = os.environ.get('SUPLEXED_PRICING_API_URL') or 'https://suplexed-pricing-api.suplexed.io'
        url = f'{suplexed_url}/{style_id}/prices'

        if url in self.local_cache_v2:
            return self.local_cache_v2[url]

        res = self.sess.get(url)
        
        if res.status_code != 200:
            raise Exception(f'[status_code={res.status_code}] Failed Fetching Suplexed Data for style_id={style_id}')

        res_json = res.json()
        # self.local_cache_v2[url] = res_json
        return res_json

    def get_pricing_data_v2_types(self, style_id: str, calling_method: str = '', marketplaces: str = '') -> SuplexedPricingResponse:
        suplexed_url = os.environ.get('SUPLEXED_PRICING_API_URL') or 'https://suplexed-pricing-api.suplexed.io'
        authorization = os.environ.get('SUPLEXED_PRICING_API_AUTHORIZATION_HEADER') or 'suplexed-admin'
        service_name = os.environ.get('SUPLEXED_SERVICE_NAME') or 'dev'

        url = f'{suplexed_url}/{style_id}/prices?marketplaces={marketplaces}'

        if url in self.local_cache_v2:
            return self.local_cache_v2[url]

        res = self.sess.get(url, headers={'authorization': authorization, 'caller': f'{service_name}:{calling_method}'})
        
        if res.status_code != 200:
            raise Exception(f'[status_code={res.status_code}] Failed Fetching Suplexed Data for style_id={style_id}')

        res_json = res.json()
        # self.local_cache_v2[url] = res_json
        return SuplexedPricingResponse(**res_json)

    def get_pricing_data_v2_size(self, style_id: str, shoe_size: str, marketplaces: str = None):
        suplexed_url = os.environ.get('SUPLEXED_PRICING_API_URL') or 'https://suplexed-pricing-api.suplexed.io'
        authorization = os.environ.get('SUPLEXED_PRICING_API_AUTHORIZATION_HEADER') or 'suplexed-admin'
        service_name = os.environ.get('SUPLEXED_SERVICE_NAME') or 'dev'

        url = f'{suplexed_url}/{style_id}/prices{f"?marketplaces={marketplaces}" if marketplaces else ""}'

        if url in self.local_cache_v2:
            pricing_response_json = self.local_cache_v2[url]
        else:
            res = self.sess.get(url, headers={'authorization': authorization, 'caller': service_name})
            
            if res.status_code != 200:
                raise Exception(f'[status_code={res.status_code}] Failed Fetching Suplexed Data for style_id={style_id}')

            pricing_response_json = res.json()
            # self.local_cache_v2[url] = pricing_response_json


    def get_pricing_data_v2_size_typed(
        self, 
        style_id: str, 
        shoe_size: str, 
        marketplaces: str,
        calling_method: str = '',
        url_key: str = None,
        shoe_name: str = ''
    ) -> SuplexedSizeResponse:
    
        suplexed_url = os.environ.get('SUPLEXED_PRICING_API_URL') or 'https://suplexed-pricing-api.suplexed.io'

        
        if style_id in self.query_overrides and url_key is not None or 'f&f' in shoe_name.lower() or 'friends' in shoe_name.lower() or 'family' in shoe_name.lower():
            query_term = url_key
        else:
            query_term = style_id

        url = f'{suplexed_url}/{query_term}/prices?marketplaces={marketplaces}'
        print(f'getting prices at {url}')

        res = self.get_pricing_data_v2_types(style_id, calling_method=calling_method, marketplaces=marketplaces)        
        # self.local_cache_v2[url] = res            


        if style_id != res.meta.styleId:
            raise Exception('Style Code Mismatch')
    
        postfix = res.meta.sizeLabel

        for size in res.listings.sizes.keys():
            formatted_size = shoe_sizes.get_stripped_size_as_float_from_string(shoe_size).replace('.0', '')
            # print(f'formatted {shoe_size} to {formatted_size}')
            if f'{size}{postfix}' == formatted_size or size == formatted_size:
                return SuplexedSizeResponse(res.meta, res.marketplaces,res.listings.sizes[size])

        raise Exception('Could not match size')

    def get_retail_price_for_style(self, style_id: str) -> int:
        suplexed_url = os.environ.get('SUPLEXED_PRICING_API_URL') or 'https://suplexed-pricing-api.suplexed.io'
        url = f'{suplexed_url}/{style_id}/prices'

        if url in self.local_cache_v2:
            pricing_response_json = self.local_cache_v2[url]
        else:
            res = self.sess.get(url)
            
            if res.status_code != 200:
                raise Exception(f'[status_code={res.status_code}] Failed Fetching Suplexed Data for style_id={style_id}')

            pricing_response_json = res.json()
            # self.local_cache_v2[url] = pricing_response_json

        return pricing_response_json['meta']['retailPrice']

if __name__ == '__main__':
    api = SuplexApi()
    typed_response = api.get_pricing_data_v2_size_typed(style_id='GZ8655', shoe_size='13')
    print(typed_response.size)
    print(typed_response.meta)
    print(typed_response.marketplaces)