import uuid
from abc import ABC, abstractclassmethod
from datetime import datetime, timedelta
from sqlalchemy.orm import Session as TSession

from suplexedlibrary.model.InventoryItem import InventoryItem
from suplexedlibrary.model.AskGoat import AskGoat
from suplexedlibrary.model.AskStockX import AskStockX
from suplexedlibrary.model.DataSyncAccount import DataSyncAccount
from suplexedlibrary.model.PurchaseOrder import PurchaseOrder
from suplexedlibrary.model.SaleOrder import SaleOrder
from suplexedlibrary.database.postgres.postgres import get_session
from suplexedlibrary.constants.Marketplace import MarketplaceEnum, Marketplace
from suplexedlibrary.model.PricingStrategy import PricingStrategy
from suplexedlibrary.constants.PricerStrategy import PricerStrategy
from suplexedlibrary.constants.ListingStatus import ListingStatus
from suplexedlibrary.kafka.KafkaSuplexedClient import KafkaSuplexedClient

class InventoryManager(ABC):
    def __init__(self, account: DataSyncAccount, database: TSession = get_session()):
        self.db = database
        self.account = account
        self.is_listing_data_healthy = True
        self.kafka_client = KafkaSuplexedClient()

    @abstractclassmethod
    def process_listings(cls):
        pass

    @abstractclassmethod
    def process_purchase_orders(cls):
        pass

    @abstractclassmethod
    def process_sale_orders(cls):
        pass

    def parse_stockx_shoe_size(self):
        pass

    def parse_goat_shoe_size(self):
        pass

    def create_inventory_and_ask_from_listing_in_db(
        self,
        shoe_name: str,
        shoe_size: any,
        shoe_sku: str,
        ask_amount: float,
        exchange: MarketplaceEnum,
        exchange_ask_id: str,
        image_url: str,
        marketplace: str,
        created_at: str,
        stx_sku_uuid: str = None,
        is_consigned: bool = None,
        status: str = None,
        product_type: str = None,
        expiry_date: str = None,
        is_expired: bool = None,
        stockx_product_url_key: str = None,
        goat_product_url_key: str = None,
    ):
        try:
            inventory_item_id = str(uuid.uuid1())

            ask = None

            if exchange == Marketplace.GOAT.value:

                existing_ask = self.db.query(AskGoat).filter(
                    AskGoat.account_id == self.account.id,
                    AskGoat.exchange_ask_id == exchange_ask_id
                ).first()

                if existing_ask and existing_ask:
                    matched_pricing_strategy = self.db.query(PricingStrategy).filter(
                        PricingStrategy.account_id == self.account.id,
                        PricingStrategy.inventory_id == existing_ask.inventory_item_id
                    ).first()
                    matched_inventory: InventoryItem = self.db.query(InventoryItem).filter(
                        InventoryItem.account_id == self.account.id,
                        InventoryItem.id == existing_ask.inventory_item_id
                    ).first()
                    matched_inventory.product_type = product_type
                    matched_inventory.goat_product_url_key = goat_product_url_key
                    matched_inventory.last_updated_time = datetime.now()

                    existing_ask.real_ask_amount = ask_amount
                    existing_ask.ask_amount = ask_amount if matched_pricing_strategy.strategy_type == 'static' else existing_ask.ask_amount                    
                    existing_ask.status = status
                    existing_ask.is_consigned = is_consigned
                    existing_ask.image_url = image_url
                    existing_ask.size = shoe_size
                    self.db.commit()
                    return
                
                print(f'creating a new inventory item for [ marketplace: {exchange} | shoe: {shoe_name} | style: {shoe_sku} | size: {shoe_size} | exchange_ask_id: {exchange_ask_id}')
                ask = AskGoat(
                    id=str(uuid.uuid1()),
                    inventory_item_id=inventory_item_id,
                    account_id=self.account.id,
                    ask_amount=ask_amount,
                    exchange=exchange,
                    exchange_ask_id=exchange_ask_id,
                    style_id=shoe_sku,
                    size=shoe_size,
                    created_at=created_at,
                    is_consigned=is_consigned,
                    status=ListingStatus.ACTIVE,
                    real_ask_amount=ask_amount
                )
            elif exchange == Marketplace.STOCKX.value:

                existing_ask = self.db.query(AskStockX).filter(
                    AskStockX.exchange_ask_id == exchange_ask_id,
                    AskStockX.account_id == self.account.id
                ).first()

                if existing_ask:
                    matched_parent_inventroy = self.db.query(InventoryItem).filter(InventoryItem.id == existing_ask.inventory_item_id).first()
                    matched_parent_inventroy.stockx_product_url_key = stockx_product_url_key
                    matched_parent_inventroy.last_updated_time = datetime.now()
                    
                    if matched_parent_inventroy.is_spec:
                        print('skipping specs')
                        return
                    matched_pricing_strategy = self.db.query(PricingStrategy).filter(
                        PricingStrategy.account_id == self.account.id,
                        PricingStrategy.inventory_id == existing_ask.inventory_item_id
                    ).first()
                    existing_ask.real_ask_amount = ask_amount
                    existing_ask.ask_amount = ask_amount if matched_pricing_strategy.strategy_type == 'static' else existing_ask.ask_amount
                    existing_ask.status = status
                    existing_ask.is_consigned = is_consigned
                    existing_ask.image_url = image_url
                    existing_ask.size = shoe_size
                    existing_ask.expiry_date = expiry_date
                    existing_ask.is_expired = is_expired
                    self.db.commit()
                    return
                
                print(f'creating a new inventory item for [ marketplace: {exchange} | shoe: {shoe_name} | style: {shoe_sku} | size: {shoe_size} | exchange_ask_id: {exchange_ask_id}')
                ask = AskStockX(
                    id=str(uuid.uuid1()),
                    inventory_item_id=inventory_item_id,
                    account_id=self.account.id,
                    ask_amount=ask_amount,
                    exchange=exchange,
                    exchange_ask_id=exchange_ask_id,
                    style_id=shoe_sku,
                    size=shoe_size,
                    created_at=created_at,
                    stx_sku_uuid=stx_sku_uuid,
                    status=ListingStatus.ACTIVE,
                    real_ask_amount=ask_amount,
                    expiry_date=expiry_date,
                    is_expired=is_expired
                )


            pricing_strategy = PricingStrategy(
                id              = str(uuid.uuid1()),
                account_id      = self.account.id,
                inventory_id    = inventory_item_id,
                expiration_date = datetime.today() + timedelta(days=14),
                strategy_type   = PricerStrategy.STATIC,
                maximize_profit = False,
                mute_goat       = False,
                mute_stockx     = False
            )

            inventory_item = InventoryItem(
                id=inventory_item_id,
                account_id=self.account.id,
                shoeName=shoe_name,
                shoeSize=shoe_size,
                shoeSku=shoe_sku,
                image_url=image_url,
                synced=False,
                last_updated_time=datetime.now(),
                product_type=product_type,
                is_spec=False,
                stockx_product_url_key=stockx_product_url_key,
                goat_product_url_key=goat_product_url_key
            )

            self.db.add(inventory_item)
            self.db.add(ask)
            self.db.add(pricing_strategy)
            self.db.commit()

        except Exception as e:
            print(e)

    def create_purchase_order_in_db(
        self,
        account_id: str,
        shoe_sku: str,
        shoe_name: str,
        shoe_size: str,
        inventory_id: str,
        exchange_ask_id: str,
        purchase_date: str,
        purchase_amount: float,
        purchase_fee: float,
        purchase_shipping: float,
        purchase_discount: float,
        marketplace: MarketplaceEnum,
        image_url: str,
        tracking_url: str,
        tracking_delivery_date: str,
        tracking_number: str,
        tracking_status: str,
        order_status: str,
        status_code: str,
        order_number: str,
        purchase_method: str,
        authentication_fee: int,
        purchase_tax: int,
        product_type: str = None,
    ) -> PurchaseOrder:
        purchase_fee = purchase_fee if purchase_fee else 0
        purchase_shipping = purchase_shipping if purchase_shipping else 0
        authentication_fee = authentication_fee if authentication_fee else 0
        purchase_discount = purchase_discount if purchase_discount else 0
        purchase_total = purchase_amount + purchase_fee + purchase_shipping + authentication_fee + purchase_discount
        
        purchase_order = PurchaseOrder(
            id=str(uuid.uuid1()),
            account_id=account_id,
            inventory_id=None,
            shoe_name=shoe_name,
            shoe_size=shoe_size,
            shoe_sku=shoe_sku,
            exchange_ask_id=exchange_ask_id,
            purchase_date=purchase_date,
            purchase_amount=purchase_amount,
            purchase_fee=purchase_fee,
            purchase_shipping=purchase_shipping,
            purchase_discount=purchase_discount,
            marketplace=marketplace.value,
            image_url=image_url,
            tracking_url=tracking_url,
            tracking_delivery_date=tracking_delivery_date,
            tracking_number=tracking_number,
            tracking_status=tracking_status,
            order_status=order_status,
            status_code=status_code,
            order_number=order_number,
            purchase_method=purchase_method,
            authentication_fee=authentication_fee,
            purchase_total=purchase_total,
            product_type=product_type,
            purchase_tax=purchase_tax
        )

        self.db.add(purchase_order)
        self.db.commit()

        purchase_formatted_amount = purchase_order.purchase_amount
        if marketplace.value == 'goat':
            purchase_formatted_amount = purchase_order.purchase_amount / 100

        return purchase_order

    def create_sale_order_in_db(self,
                                account_id: str,
                                shoe_sku: str,
                                shoe_name: str,
                                shoe_size: str,
                                inventory_id: str,
                                exchange_ask_id: str,
                                sold_date: str,
                                sold_amount: float,
                                sold_payout: float,
                                sold_transaction_fee: float,
                                sold_payment_processing_fee: float,
                                sold_shipping: float,
                                sold_discount: float,
                                marketplace: MarketplaceEnum,
                                image_url: str,
                                order_status: str,
                                tracking_url: str,
                                tracking_delivery_date: str,
                                tracking_number: str,
                                tracking_status: str,
                                status_code: str,
                                order_number: str,
                                payout_destination: str,
                                listing_creation_date: str,
                                shipping_label_url: str,
                                is_sale_confirmed: bool
                                ) -> SaleOrder:
        print(f'creating a new sale order for inventory_id={inventory_id}, style_id={shoe_sku}, account_id={account_id}')

        sale_order = SaleOrder(
            id=str(uuid.uuid1()),
            account_id=account_id,
            inventory_id=inventory_id,
            shoe_name=shoe_name,
            shoe_size=shoe_size,
            shoe_sku=shoe_sku,
            exchange_ask_id=exchange_ask_id,
            sold_date=sold_date,
            sold_amount=sold_amount,
            sold_payout=sold_payout,
            sold_transaction_fee=sold_transaction_fee,
            sold_payment_processing_fee=sold_payment_processing_fee,
            sold_shipping=sold_shipping,
            sold_discount=sold_discount,
            marketplace=marketplace.value,
            image_url=image_url,
            order_status=order_status,
            tracking_url=tracking_url,
            tracking_delivery_date=tracking_delivery_date,
            tracking_number=tracking_number,
            tracking_status=tracking_status,
            status_code=status_code,
            order_number=order_number,
            payout_destination=payout_destination,
            listing_creation_date=listing_creation_date,
            shipping_label_url=shipping_label_url,
            is_sale_confirmed=is_sale_confirmed
        )

        self.db.add(sale_order)
        self.db.commit()

        return sale_order

    def update_sale_order_in_db(
        self,
        sale_order: SaleOrder,
        order_status: str,
        tracking_url: str,
        tracking_delivery_date: str,
        tracking_number: str,
        tracking_status: str,
        status_code: str,
        order_number: str,
        sold_tax: int,
        sold_amount:int = None,
        sold_payout:int = None,
        sold_transaction_fee:int = None,
        sold_payment_processing_fee:int = None,
        sold_shipping:int = None,
        shipping_label_url: str = None,
        is_sale_confirmed: bool = None,
    ) -> SaleOrder:
        print(f'updating sale order for sale_order={sale_order.id}, style_id={sale_order.shoe_sku}, account_id={sale_order.account_id}')
        sale_order.order_status = order_status
        sale_order.tracking_url = tracking_url
        sale_order.tracking_delivery_date = tracking_delivery_date
        sale_order.tracking_number = tracking_number
        sale_order.tracking_status = tracking_status
        sale_order.status_code = status_code
        sale_order.order_number = order_number
        sale_order.shipping_label_url = shipping_label_url
        sale_order.is_sale_confirmed = is_sale_confirmed
        sale_order.sold_tax = sold_tax

        if sold_amount:
            sale_order.sold_amount = sold_amount
        if sold_payout:
            sale_order.sold_payout = sold_payout
        if sold_transaction_fee:
            sale_order.sold_transaction_fee = sold_transaction_fee
        if sold_payment_processing_fee:
            sale_order.sold_payment_processing_fee = sold_payment_processing_fee
        if sold_shipping:
            sale_order.sold_shipping = sold_shipping
            
        self.db.commit()
        return sale_order

    def update_purchase_order_in_db(
        self,
        purchase_order: PurchaseOrder,
        order_status: str,
        tracking_url: str,
        tracking_delivery_date: str,
        tracking_number: str,
        tracking_status: str,
        status_code: str,
        order_number: str,
        purchase_amount: int,
        purchase_shipping: int,
        purchase_fee: int,
        purchase_tax: int
    ):
        print(f'updating purchase order for sale_order={purchase_order.id}, style_id={purchase_order.shoe_sku}, account_id={purchase_order.account_id}')
        purchase_order.order_status = order_status
        purchase_order.tracking_url = tracking_url
        purchase_order.tracking_delivery_date = tracking_delivery_date
        purchase_order.tracking_number = tracking_number
        purchase_order.tracking_status = tracking_status
        purchase_order.status_code = status_code
        purchase_order.order_number = order_number
        purchase_order.purchase_amount = purchase_amount
        purchase_order.purchase_shipping = purchase_shipping
        purchase_order.purchase_fee = purchase_fee
        purchase_order.purchase_tax = purchase_tax
        self.db.commit()
        return purchase_order

    def create_sale_event(self, sale_event: dict):
        self.kafka_client.send_message(KafkaSuplexedClient.NEW_SALE_EVENT_TOPIC, sale_event)

    def update_sale_event(self, update_sale_event: dict):
        self.kafka_client.send_message(KafkaSuplexedClient.UPDATE_SALE_EVENT_TOPIC, update_sale_event)