from datetime import date, datetime
from uuid import uuid4
from requests import session, Session
from typing import List, Set
from sqlalchemy.orm import Session as TSession
from concurrent.futures import ThreadPoolExecutor

from suplexedlibrary.constants.OrderStatusStockx import OrderStatusStockx
from suplexedlibrary.constants.UserEventType import UserEventType
from suplexedlibrary.inventory.InventoryManager import InventoryManager
from suplexedlibrary.model.StockxAccount import StockxAccount
from suplexedlibrary.model.SaleOrder import SaleOrder
from suplexedlibrary.api.StockxApi import StockXApi
from suplexedlibrary.model.UserEvent import UserEvent
from suplexedlibrary.util.stockx_util import Fees
from suplexedlibrary.model.DataSyncAccount import DataSyncAccount
from suplexedlibrary.model.InventoryItem import InventoryItem
from suplexedlibrary.model.AskStockX import AskStockX
from suplexedlibrary.model.PurchaseOrder import PurchaseOrder
from suplexedlibrary.database.postgres.postgres import get_session
from suplexedlibrary.constants.Marketplace import Marketplace
from suplexedlibrary.constants.Marketplace import Marketplace
from suplexedlibrary.discord import hooks
from suplexedlibrary.exceptions import FailedFetchingListings
from suplexedlibrary.constants.ListingStatus import ListingStatus

class InventoryManagerStockX(InventoryManager):
    def __init__(self, account: DataSyncAccount, stockx_account: StockxAccount, session: Session, database: TSession = get_session()):
        super().__init__(account, database)
        self.stockx_account = stockx_account
        self.stockx_api = StockXApi(self.stockx_account, skip_login=False)
        self.listings = []
        self.goat_storage_listing = []
        self.purchase_orders_pending = []
        self.purchase_orders_history = []
        self.sale_orders_pending = []
        self.sale_orders_history = []
        self.sales_order_history_gql = []


    def get_and_set_purchase_order_history(self):
        try:
            self.purchase_orders_history = self.stockx_api.get_all_buying(buy_type='history')
        except Exception as e:
            print(e)
            raise e

    def get_and_set_purchase_order_pending(self):
        try:
            self.purchase_orders_pending = self.stockx_api.get_all_buying(buy_type='pending')
        except Exception as e:
            print(e)
            raise e

    def get_and_set_listings(self):
        try:
            self.listings = self.stockx_api.get_all_sales(sale_type='current')
        except FailedFetchingListings as e:
            print(f'caught listing fetch failure: {e}')
            self.is_listing_data_healthy = False
            raise e
        except Exception as e:
            print(e)
            self.is_listing_data_healthy = False
            raise e

    def get_and_set_sale_orders_pending(self):
        try:
            self.sale_orders_pending = self.stockx_api.get_all_sales(sale_type='pending')
        except Exception as e:
            print(e)
            raise e

    def get_and_set_sale_orders_history(self, is_full_scrape: bool):
        try:
            self.sale_orders_history = self.stockx_api.get_all_sales(sale_type='history', is_full_scrape=is_full_scrape)
        except Exception as e:
            print(e)
            raise e


    def get_and_set_sales_orders_history_gql(self):
        try:
            self.sales_order_history_gql = self.stockx_api.get_sales_history_gql()
        except Exception as e:
            print(e)
            raise e

    def hydrate_orders(self):
        hydration_tasks = [
            self.get_and_set_purchase_order_history,
            self.get_and_set_purchase_order_pending,
            self.get_and_set_sale_orders_history,
            self.get_and_set_sale_orders_pending
        ]
        with ThreadPoolExecutor(5) as executor:
            [executor.submit(task) for task in hydration_tasks]

    def hydrate_account(self):     
        hydration_tasks = [
            self.get_and_set_listings,
            self.get_and_set_purchase_order_history,
            self.get_and_set_purchase_order_pending,
            self.get_and_set_sale_orders_history,
            self.get_and_set_sale_orders_pending
        ]

        with ThreadPoolExecutor(5) as executor:
            [executor.submit(task) for task in hydration_tasks]

    def process_listings(self):

        for listing in self.listings:
            self.create_inventory_and_ask_from_listing_in_db(
                shoe_name=listing['product']['title'],
                shoe_size=listing['product']['shoeSize'],
                shoe_sku=listing['product']['styleId'],
                ask_amount=listing['amount'],
                exchange=Marketplace.STOCKX.value,
                exchange_ask_id=listing['chainId'],
                image_url=listing['product']['media']['imageUrl'],
                marketplace='stockx',
                created_at=listing['createdAt'],
                stx_sku_uuid=listing['skuUuid'],
                status='active' if listing['state'] == 400 else 'inactive',
                product_type=listing['product']['contentGroup'],
                expiry_date=listing['expiresAt'],
                is_expired=False if 'expiresAt' not in listing else datetime.strptime(listing['expiresAt'].split('+')[0], "%Y-%m-%dT%H:%M:%S") < datetime.now(),
                stockx_product_url_key=listing['product']['urlKey']
            )

    def process_purchase_orders(self, limit: int = 2):
        print(f'processing purchase orders for {self.stockx_account.email}')

        purchase_orders = self.purchase_orders_history + self.purchase_orders_pending

        for order in purchase_orders:

            chain_id = order['chainId']
            order_number = order['orderNumber']
            matched_purchase_order = self.db.query(PurchaseOrder).filter(
                PurchaseOrder.order_number == order_number,
                PurchaseOrder.account_id == self.account.id
            ).first()

            transaction_fee = None
            authentication_fee = None
            shipping = None
            purchase_method = None
            if not matched_purchase_order: # If we have not recorded this sale object yet, create one for it

                try:
                    detailed_order = self.stockx_api.get_order_details(chain_id)
                    if detailed_order:
                        print('success fetching order details')
                        subtotal = detailed_order['PortfolioItem']['Pricing']['subtotal']
                        payout_usd = detailed_order['PortfolioItem']['Pricing']['totalUSD']
                        adjustments = detailed_order['PortfolioItem']['Pricing']['adjustments']
                        if len(adjustments) == 3:
                            shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']
                            transaction_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                            authentication_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][2]['value']
                        elif len(adjustments) == 2:
                            shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']  
                            authentication_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                            transaction_fee = 0
                        purchase_method = detailed_order['PortfolioItem']['Payment']['descriptor']
                    else:
                        print('no detailed order info found')
                except Exception as e:
                    # hooks.write_to_error_channel(self.account.email, f'{e}: failed getting order details')
                    print('failed getting order details')
                    print(e)
                    continue

                self.create_purchase_order_in_db(
                    account_id=self.account.id,
                    shoe_sku=order['product']['styleId'],
                    shoe_name=order['product']['title'],
                    shoe_size=order['product']['shoeSize'],
                    inventory_id=None,
                    exchange_ask_id=order['chainId'],
                    purchase_date=order['matchedWithDate'],
                    purchase_amount=order['amount'],
                    purchase_fee=transaction_fee,
                    purchase_shipping=shipping,
                    purchase_discount=0,
                    marketplace=Marketplace.STOCKX,
                    image_url=order['product']['media']['imageUrl'],
                    tracking_url=order['Tracking']['trackUrl'] if 'Tracking' in order else None,
                    tracking_delivery_date=order['Tracking']['deliveryDate'] if 'Tracking' in order else None,
                    tracking_number=order['Tracking']['number'] if 'Tracking' in order else None,
                    tracking_status=order['Tracking']['status'] if 'Tracking' in order else None,
                    order_status=order['text'],
                    status_code=order['state'],
                    order_number=order['orderNumber'],
                    purchase_method=purchase_method,
                    authentication_fee=authentication_fee,
                    product_type=order['product']['contentGroup'],
                    purchase_tax=0
                )
            else:
                if not matched_purchase_order.purchase_amount:
                    try:
                        detailed_order = self.stockx_api.get_order_details(chain_id)
                        if detailed_order:
                            print('success fetching order details')
                            subtotal = detailed_order['PortfolioItem']['Pricing']['subtotal']
                            payout_usd = detailed_order['PortfolioItem']['Pricing']['totalUSD']
                            adjustments = detailed_order['PortfolioItem']['Pricing']['adjustments']
                            if len(adjustments) == 3:
                                shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']
                                transaction_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                                authentication_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][2]['value']
                            elif len(adjustments) == 2:
                                shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']  
                                authentication_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                                transaction_fee = 0
                            purchase_method = detailed_order['PortfolioItem']['Payment']['descriptor']
                        else:
                            print('no detailed order info found')
                    except Exception as e:
                        # hooks.write_to_error_channel(self.account.email, f'{e}: failed getting order details')
                        print('failed getting order details')
                        print(e)
                        continue     
                self.update_purchase_order_in_db(
                    matched_purchase_order,
                    order_status=order['text'],
                    tracking_url=order['Tracking']['trackUrl'] if 'Tracking' in order else None,
                    tracking_delivery_date=order['Tracking']['deliveryDate'] if 'Tracking' in order else None,
                    tracking_number=order['Tracking']['number'] if 'Tracking' in order else None,
                    tracking_status=order['Tracking']['status'] if 'Tracking' in order else None,
                    status_code=order['state'],
                    order_number=order['orderNumber'],
                    purchase_amount=order['amount'],
                    purchase_shipping=shipping,
                    purchase_fee=transaction_fee,
                    purchase_tax=0,
                )                

    def process_sale_orders(self, limit: int = 2):

        sale_orders = self.sale_orders_pending + self.sale_orders_history

        for i, order in enumerate(sale_orders):
            try:
                chain_id = order['chainId']
                matched_sale_order = self.db.query(SaleOrder).filter(
                    SaleOrder.exchange_ask_id == chain_id,
                    SaleOrder.account_id == self.account.id
                ).first()

                if not matched_sale_order: # If we have not recorded this sale object yet, create one for it
                    try:
                        detailed_order = self.stockx_api.get_order_details(chain_id)
                        if detailed_order:
                            subtotal = detailed_order['PortfolioItem']['Pricing']['subtotal']
                            payout_usd = detailed_order['PortfolioItem']['Pricing']['totalUSD']
                            transaction_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']
                            payment_processing_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                            shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][2]['value']
                            payout_account = detailed_order['PortfolioItem']['Payout']['descriptor']
                        else:
                            print('no detailed order found')
                            subtotal = None
                            payout_usd = None
                            transaction_fee = None
                            payment_processing_fee = None
                            shipping = None
                            payout_account = None
                        
                    except Exception as e:
                        # hooks.write_to_error_channel(self.account.email, f'{e}: failed getting order details')
                        print('failed getting order details')
                        print(e)
                        continue

                    self.create_sale_event({
                        'account_id': self.account.id,
                        'shoe_sku': order['product']['styleId'],
                        'shoe_name': order['product']['title'],
                        'shoe_size': order['product']['shoeSize'],
                        'exchange_ask_id': chain_id,
                        'sold_date': order['matchedWithDate'],
                        'sold_amount': subtotal,
                        'sold_payout': payout_usd,
                        'sold_transaction_fee': transaction_fee,
                        'sold_payment_processing_fee': payment_processing_fee,
                        'sold_shipping': shipping,
                        'sold_discount': 0,
                        'sold_tax': 0,
                        'marketplace': Marketplace.STOCKX.value,
                        'image_url': order['product']['media']['imageUrl'],
                        'order_status': order['text'],
                        'tracking_url': order['Tracking']['trackUrl'] if 'Tracking' in order else None,
                        'tracking_delivery_date': order['Tracking']['deliveryDate'] if 'Tracking' in order else None,
                        'tracking_number': order['Tracking']['number'] if 'Tracking' in order else None,
                        'tracking_status': order['Tracking']['status'] if 'Tracking' in order else None,
                        'status_code': order['state'],
                        'order_number': order['orderNumber'],
                        'payout_destination': payout_account,
                        'listing_creation_date': order['createdAt'],
                        'shipping_label_url': order['Tracking']['labelUrl'] if 'Tracking' in order and 'labelUrl' in order['Tracking'] else None,
                        'is_sale_confirmed': None                        
                    })


                else: # Otherwise update the status of the sale order existing in the database
                    
                    if not(matched_sale_order.sold_amount or matched_sale_order.sold_payout or matched_sale_order.sold_payment_processing_fee):
                        detailed_order = self.stockx_api.get_order_details(chain_id)
                    else:
                        detailed_order = None
                        subtotal = matched_sale_order.sold_amount
                        payout_usd = matched_sale_order.sold_payout
                        transaction_fee = matched_sale_order.sold_transaction_fee
                        payment_processing_fee = matched_sale_order.sold_payment_processing_fee
                        shipping = matched_sale_order.sold_shipping
                        payout_account = matched_sale_order.payout_destination
            
                    if detailed_order:
                        subtotal = detailed_order['PortfolioItem']['Pricing']['subtotal']
                        payout_usd = detailed_order['PortfolioItem']['Pricing']['totalUSD']
                        transaction_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']
                        payment_processing_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                        shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][2]['value']
                        payout_account = detailed_order['PortfolioItem']['Payout']['descriptor']

                    self.update_sale_event({
                        'marketplace': 'stockx',
                        'sale_order_id': chain_id,
                        'account_id': self.account.id,
                        'order_status': order['text'],
                        'tracking_url': order['Tracking']['trackUrl'] if 'Tracking' in order else None,
                        'tracking_delivery_date': order['Tracking']['deliveryDate'] if 'Tracking' in order else None,
                        'tracking_number': order['Tracking']['number'] if 'Tracking' in order else None,
                        'tracking_status': order['Tracking']['status'] if 'Tracking' in order else None,
                        'status_code': order['state'],
                        'order_number': order['orderNumber'],
                        'sold_amount': subtotal,
                        'sold_payout': payout_usd,
                        'sold_tax': 0,
                        'sold_transaction_fee': transaction_fee,
                        'sold_payment_processing_fee': payment_processing_fee,
                        'sold_shipping': shipping,
                        'shipping_label_url': order['Tracking']['labelUrl'] if 'Tracking' in order and 'labelUrl' in order['Tracking'] else None,
                        'is_sale_confirmed': None
                    })

            except Exception as e:
                print(f'e occurred while processing sale for {self.account.email}')
                print(e)

    def process_sale_orders_gql(self):
        for order in self.sales_order_history_gql:

            matched_order = self.db.query(SaleOrder).filter(SaleOrder.exchange_ask_id == order['id'], SaleOrder.account_id == self.account.id).first()

            if not matched_order:
                print(f'no order found in DB for {order["id"]}')
                try:
                    detailed_order = self.stockx_api.get_order_details(order['id'])
                    if detailed_order:
                        subtotal = detailed_order['PortfolioItem']['Pricing']['subtotal']
                        payout_usd = detailed_order['PortfolioItem']['Pricing']['totalUSD']
                        transaction_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']
                        payment_processing_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                        shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][2]['value']
                        payout_account = detailed_order['PortfolioItem']['Payout']['descriptor']
                    else:
                        print('no detailed order found')
                        subtotal = None
                        payout_usd = None
                        transaction_fee = None
                        payment_processing_fee = None
                        shipping = None
                        payout_account = None

                except Exception as e:
                    # hooks.write_to_error_channel(self.account.email, f'{e}: failed getting order details')
                    print('failed getting order details')
                    print(e)
                    subtotal = None
                    payout_usd = None
                    transaction_fee = None
                    payment_processing_fee = None
                    shipping = None
                    payout_account = None                    

                self.create_sale_event({
                    'account_id': self.account.id,
                    'shoe_sku': order['productVariant']['product']['styleId'],
                    'shoe_name': order['productVariant']['product']['model'] + ' ' + order['productVariant']['product']['name'],
                    'shoe_size': order['productVariant']['traits']['size'],
                    'exchange_ask_id': order['chainId'],
                    'sold_date': order['soldOn'],
                    'sold_amount': subtotal,
                    'sold_payout': payout_usd,
                    'sold_transaction_fee': transaction_fee,
                    'sold_payment_processing_fee': payment_processing_fee,
                    'sold_shipping': shipping,
                    'sold_discount': 0,
                    'sold_tax': 0,
                    'marketplace': Marketplace.STOCKX.value,
                    'image_url': order['productVariant']['product']['media']['imageUrl'],
                    'order_status': order['status'][1],
                    'tracking_url': None, #order['Tracking']['trackUrl'] if 'Tracking' in order else None,
                    'tracking_delivery_date': None, # order['Tracking']['deliveryDate'] if 'Tracking' in order else None,
                    'tracking_number': None, #order['Tracking']['number'] if 'Tracking' in order else None,
                    'tracking_status': None, #order['Tracking']['status'] if 'Tracking' in order else None,
                    'status_code': order['state'],
                    'order_number': order['orderNumber'],
                    'payout_destination': payout_account,
                    'listing_creation_date': None, # order['createdAt'],
                    'shipping_label_url': None, #order['Tracking']['labelUrl'] if 'Tracking' in order and 'labelUrl' in order['Tracking'] else None,
                    'is_sale_confirmed': None                        
                })          

            else:
                if not matched_order.sold_amount:
                    try:
                        detailed_order = self.stockx_api.get_order_details(order['id'])
                        if detailed_order:
                            matched_order.sold_amount = detailed_order['PortfolioItem']['Pricing']['subtotal']
                            matched_order.sold_payout = detailed_order['PortfolioItem']['Pricing']['totalUSD']
                            matched_order.sold_transaction_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][0]['value']
                            matched_order.sold_payment_processing_fee = detailed_order['PortfolioItem']['Pricing']['adjustments'][1]['value']
                            matched_order.sold_shipping = detailed_order['PortfolioItem']['Pricing']['adjustments'][2]['value']
                            matched_order.payout_destination = detailed_order['PortfolioItem']['Payout']['descriptor']

                    except:
                        print('could not get the sale details')
                # matched_order.order_status = order['status'][1]
                matched_order.status_code = order['state']
                matched_order.order_status = order['status'][1]
            
            self.db.commit()

    @staticmethod
    def process_missing_listings_static(listings: List[AskStockX], recent_sales: List[AskStockX], account_id: str, stockx_api: StockXApi, created_listings: Set[str]):
        db_session = get_session()
        all_asks = db_session.query(AskStockX).filter(
            AskStockX.account_id == account_id
        ).all()

        ask_set = {}
        exchange_ask_set = {}

        for ask in all_asks:
            ask_set[ask.exchange_ask_id] = ask
        
        original_size = len(ask_set)

        for ask in listings + recent_sales:
            ask_set.pop(ask['chainId'], None)
            exchange_ask_set[ask['chainId']] = ask

        for stale_ask in ask_set:
            if ask_set[stale_ask].status == ListingStatus.PENDING:
                print(f'this ask is still pending do not delete: {stale_ask}')
                continue

            if stale_ask in created_listings:
                print(f'this ask was just created, do not delete: {stale_ask}')
                continue

            inventory_item = db_session.query(InventoryItem).filter(InventoryItem.id == ask_set[stale_ask].inventory_item_id).first()

            print(f'removing stale ask:{stale_ask}, [style:{ask_set[stale_ask].style_id} | size:{ask_set[stale_ask].size}]')
            user_event = UserEvent(
                id                                 = str(uuid4()),
                account_id                         = account_id,
                inventory_id                       = inventory_item.id,       
                sale_order_id                      = None,
                purchase_order_id                  = None,               
                event_type                         = UserEventType.LISTING_REMOVE_OFF_ASSIST,
                event_title                        = f'Listing removed off of sync platform for {inventory_item.shoeName} size {inventory_item.shoeSize} on StockX',
                event_text                         = f'Listing removed off of sync platform for {inventory_item.shoeName} size {inventory_item.shoeSize} on StockX',
                event_link                         = None,
                event_time                         = datetime.now(),
                event_image                        = inventory_item.image_url                
            )
            db_session.add(user_event)
            db_session.delete(ask_set[stale_ask])
            db_session.commit()

if __name__ == '__main__':
    db = get_session()
    account = db.query(DataSyncAccount).filter(DataSyncAccount.email == 'daniel.revie1@gmail.com').first()
    # im = InventoryManagerStockX(account, account.stockxAccount[0], session())
    api = StockXApi(account.stockxAccount[0])
    api.get_gql_schema()
    # im.get_and_set_sales_orders_history_gql()

