from datetime import datetime
from uuid import uuid4
from requests import session
from typing import List
from sqlalchemy import or_, and_
from pprint import pprint

from suplexedlibrary.constants.UserEventType import UserEventType
from suplexedlibrary.inventory.InventoryManagerGoat import InventoryManagerGoat
from suplexedlibrary.constants.ListingStatus import ListingStatus
from suplexedlibrary.api.GoatApi import GoatApi
from suplexedlibrary.pricing.Pricer import Pricer
from suplexedlibrary.database.postgres.postgres import get_session
from suplexedlibrary.model.DataSyncAccount import DataSyncAccount
from suplexedlibrary.model.InventoryItem import InventoryItem
from suplexedlibrary.model.ResellPricing import ResellPricing
from suplexedlibrary.model.InventorySync import InventorySync
from suplexedlibrary.model.GoatAccount import GoatAccount
from suplexedlibrary.model.AskGoat import AskGoat
from suplexedlibrary.model.UserEvent import UserEvent
from suplexedlibrary.exceptions import DeleteListingFailedException, FailedFetchingListings, CreateListingException
from suplexedlibrary.util import shoe_sizes


def myround(x, base=5):
    return base * round(x/base)


def inventory_filter(inventory_item: InventoryItem) -> bool:
    if inventory_item.pricingStrategy.mute_goat == True:
        print('this strategy muted goat')
        return False

    return True


class GoatPricer(Pricer):
    ASK_ID = 'goatAskId'

    def __init__(self, account: DataSyncAccount, goat_account: GoatAccount):
        self.asks = {}
        self.listings = []
        self.db_session = get_session()
        super().__init__(account)
        self.api = GoatApi(goat_account, skip_login=False)
        self.hydrate_ask_map()
        self.created_listings = set()
        print('Init goat pricer complete')


    def process_missing_asks(self):
        if self.is_listing_data_healthy:
            InventoryManagerGoat.process_missing_listings_static(self.listings, self.account.id, self.api, self.created_listings)
        else:
            print('listing data is not healthy -- skipping missing asks')


    def hydrate_ask_map(self):
        asks = []
        try:
            asks = self.api.get_listings()
        except FailedFetchingListings as e:
            print(f'caught listing fetch failure: {e}')
            self.is_listing_data_healthy = False
        except Exception as e:
            print(e)
            self.is_listing_data_healthy = False

        goat_storage = []
        try:
            goat_storage = self.api.get_goat_storage_listings()
        except FailedFetchingListings as e:
            print(f'caught listing fetch failure: {e}')
            self.is_listing_data_healthy = False
        except Exception as e:
            print(e)
            self.is_listing_data_healthy = False
    
        all_listings = goat_storage + asks
        self.listings = all_listings
        for ask in all_listings:
            self.asks[ask['id']] = ask

    def is_ask_inactive(self, inventory_item: InventoryItem):
        return inventory_item.askGoat.status == ListingStatus.INACTIVE

    def is_ask_deleted(self, inventory_item: InventoryItem):
        return inventory_item.askGoat.status == ListingStatus.DELETE

    def does_ask_need_update(self, inventory_item: InventoryItem) -> bool:
        matched_ask = self.asks[inventory_item.askGoat.exchange_ask_id] if inventory_item.askGoat.exchange_ask_id in self.asks else None

        if not matched_ask:
            raise Exception(f'No ask found for {inventory_item.askGoat.exchange_ask_id}')

        return matched_ask['priceCents'] / 100 != self.get_ask_amount(inventory_item)

    def is_at_bottom_market(self, inventory_item: InventoryItem, lowest_ask: int) -> bool:
        current_ask = self.get_ask_amount(inventory_item)
        current_ask = inventory_item.askGoat.ask_amount
        exchange_ask = self.asks[inventory_item.askGoat.exchange_ask_id]['priceCents'] / 100
        if lowest_ask >= exchange_ask:
            print('\tLowest ask = ' + str(lowest_ask))
            print('\tCurrent ask = ' + str(current_ask))
            return True

        return False

    def is_listed_on_exchange(self, shoe: InventoryItem) -> bool:
        if not shoe.askGoat:
            return False
        return shoe.askGoat.exchange_ask_id in self.asks

    def is_at_floor(self, inventory_item: InventoryItem) -> bool:

        ask: AskGoat = inventory_item.askGoat
        if ask == None:
            raise Exception('This ask should not be None')

        if (int(self.asks[ask.exchange_ask_id]['priceCents'] / 100)) - 1 < ask.ask_amount:
            print('ask of ' + str(ask.ask_amount) +
                  ' is already at the floor of ' + str(self.asks[ask.exchange_ask_id]['priceCents'] / 100))
            return True

        return False

    def is_bottom_market_our_shoe(self, inventory_item: InventoryItem, bottom_market: int) -> bool:
        all_asks = self.db_session.query(AskGoat).filter(
            AskGoat.size == inventory_item.shoeSize,
            AskGoat.style_id == inventory_item.shoeSku,
            AskGoat.account_id == inventory_item.account_id
        ).all()
        matched_asks = list(filter(lambda ask_key : (self.asks[ask_key]['productTemplate']['sku'].replace(' ', '-') == inventory_item.shoeSku and self.asks[ask_key]['productTemplate']['sku'].replace(' ', '-')), self.asks))

        for ask in matched_asks:
            if int(self.asks[ask]['priceCents']) / 100 == bottom_market and self.asks[ask]['slug']:
                return True

        return False

    def is_ask_corrupt(self, shoe: InventoryItem):
        if shoe.askGoat == None:
            return True
        return False

    def list_inventory_item_ask(self, inventory_item: InventoryItem, amount: int = None, start_price: int = None):
        ask_amount: int = inventory_item.askGoat.ask_amount

        if amount: # Amount override for floor pricing
            ask_amount = amount

        if start_price:
            ask_amount = start_price

        goat_obj_url = self.suplex_api.get_goat_product_url(inventory_item.shoeSku)
        product_template_id = self.api.get_product_template_id(goat_obj_url)

        if not product_template_id:
            print(f'could not find template id for: {inventory_item.shoeSku}')
            return

        try:
            goat_ask_id = self.api.create_new_own(
                inventory_id=inventory_item.id,
                amount=self.api.get_goat_amount_cents(ask_amount),
                product_template_id=product_template_id,
                size=shoe_sizes.get_stripped_size_as_float_from_string(inventory_item.shoeSize),
                sell_floor=self.api.get_goat_amount_cents(ask_amount)
            )
        except CreateListingException as e:
            print(e)
            event = UserEvent(
                id                                 = str(uuid4()),
                account_id                         = self.account.id,
                inventory_id                       = inventory_item.id,       
                sale_order_id                      = None,
                purchase_order_id                  = None,
                event_type                         = UserEventType.ERROR,
                event_title                        = f'Failed creating new ask for {inventory_item.shoeName} size {inventory_item.shoeSize} on GOAT',
                event_text                         = f'Failed creating a new ask for ${ask_amount:,}.00 on GOAT.',
                event_link                         = None,
                event_time                         = datetime.now(),
                event_image                        = inventory_item.image_url
            )
            return


        event = UserEvent(
            id                                 = str(uuid4()),
            account_id                         = self.account.id,
            inventory_id                       = inventory_item.id,       
            sale_order_id                      = None,
            purchase_order_id                  = None,
            event_type                         = UserEventType.CREATED_NEW_ASK,
            event_title                        = f'Creating new ask for {inventory_item.shoeName} size {inventory_item.shoeSize} on GOAT',
            event_text                         = f'You have posted a new listing for sale on GOAT.',
            event_link                         = None,
            event_time                         = datetime.now(),
            event_image                        = inventory_item.image_url
        )
        self.db_session.add(event)


        self.created_listings.add(goat_ask_id)
        inventory_item.askGoat.status = ListingStatus.ACTIVE

        if inventory_item.askGoat:
            inventory_item.askGoat.exchange_ask_id = goat_ask_id
            inventory_item.askGoat.real_ask_amount = ask_amount
        else:
            ask = AskGoat(
                id=str(uuid4()),
                inventory_item_id=inventory_item.id,
                ask_amount=ask_amount,
                real_ask_amoun=ask_amount,
                exchange='goat',
                exchange_ask_id=goat_ask_id,
                style_id=inventory_item.shoeSku,
                size=float(inventory_item.shoeSize),
                status=ListingStatus.ACTIVE
            )
            self.db_session.add(ask)

        self.db_session.commit()

    def update_inventory_item_ask(self, inventory_item: InventoryItem, new_ask: int, is_floor_update: bool):
        self.api.update_own(
            amount=self.api.get_goat_amount_cents(new_ask),
            slug=inventory_item.askGoat.exchange_ask_id,
            sell_floor=self.api.get_goat_amount_cents(new_ask)
        )
        event_type = None

        if is_floor_update:
            event_type = UserEventType.SELL_FLOOR_PRICE_UPDATE
        else:
            event_type = UserEventType.STATIC_PRICE_UPDATE

        event = UserEvent(
            id                                 = str(uuid4()),
            account_id                         = self.account.id,
            inventory_id                       = inventory_item.id,       
            sale_order_id                      = None,
            purchase_order_id                  = None,
            event_type                         = event_type,
            event_title                        = f'Updating ask for {inventory_item.shoeName} size {inventory_item.shoeSize} on GOAT',
            event_text                         = f'Your listing price has been automatically updated from ${inventory_item.askGoat.real_ask_amount:,}.00 to ${new_ask:,}.00 on GOAT.',
            event_link                         = None,
            event_time                         = datetime.now(),
            event_image                        = inventory_item.image_url
        )
        self.db_session.add(event)
        self.db_session.commit()
            
    def delete_inventory_item_ask(self, inventory_item: InventoryItem):
        print(f'deleting exchange ask and ask object {inventory_item.id}')
        try:
            self.api.delete_own(inventory_item.askGoat.exchange_ask_id)
        except DeleteListingFailedException as e:
            print(f'caught a listing deletion failure: {str(e)}')
            return
        except Exception as e:
            print(e)
            return

        self.db_session.query(AskGoat).filter(AskGoat.id == inventory_item.askGoat.id).delete()
        # self.db_session.delete(inventory_item.askGoat)
        self.db_session.commit()

    def delete_exchange_ask(self, inventory_item: InventoryItem):
        print(f'deleting exchange ask {inventory_item.id}')
        self.api.delete_own(inventory_item.askGoat.exchange_ask_id)
        inventory_item.askGoat.exchange_ask_id = None
        self.db_session.commit()

    def remove_all_asks(self):
        inventory_item = list(
            filter(lambda item: item.askGoat, self.account.inventory))
        for item in inventory_item:
            id = self.api.delete_own(item.askGoat.exchange_ask_id)
            self.db_session.query(AskGoat).filter(AskGoat.exchange_ask_id == id).delete()
            self.db_session.commit()

    def get_lowest_ask(self, resell_pricing: ResellPricing):
        return resell_pricing.goat_ask

    def get_ask_amount(self, inventory_item: InventoryItem) -> int:
        return inventory_item.askGoat.ask_amount

    def get_inventory(self) -> List[InventoryItem]:
        inventory_items = self.db_session.query(InventoryItem).filter(
            InventoryItem.account_id == self.account.id,
            InventoryItem.askGoat != None,
            InventoryItem.is_spec == False
        ).all()

        def filter_valid_goat_inventory_items(inventory_item: InventoryItem) -> bool:
            return inventory_item.askGoat and inventory_item.askGoat.ask_amount != None
        
        inventory_items = list(filter(lambda x : filter_valid_goat_inventory_items(x), inventory_items))
        
        return inventory_items

    def is_lowest_ask_calculated(self, resell_pricing: ResellPricing):
        return resell_pricing.is_goat_ask_calculated

    def has_active_listing_been_removed_from_the_marketplace(self, inventory_item: InventoryItem):
        ask: AskGoat = inventory_item.askGoat
        if ask == None:
            raise Exception('This ask should not be None')

        if ask.status == ListingStatus.ACTIVE and ask.exchange_ask_id not in self.asks:
            print('The user must have deleted this on the goat side, lets remove it from our db for sanity')

            user_event = UserEvent(
                id                                 = str(uuid4()),
                account_id                         = self.account.id,
                inventory_id                       = inventory_item.id,       
                sale_order_id                      = None,
                purchase_order_id                  = None,               
                event_type                         = UserEventType.LISTING_REMOVE_OFF_ASSIST,
                event_title                        = f'Active listing removed off of sync platform for {inventory_item.shoeName} size {inventory_item.shoeSize} on GOAT',
                event_text                         = f'Active listing removed off of sync platform for {inventory_item.shoeName} size {inventory_item.shoeSize} on GOAT',
                event_link                         = None,
                event_time                         = datetime.now(),
                event_image                        = inventory_item.image_url                
            )
            self.db_session.add(user_event)         
            self.db_session.delete(ask)
            inventory_item.synced = False
            self.db_session.commit()
            return True

        return False

    def is_listing_expired(self, inventory_item: InventoryItem):
        pass

    def mark_listing_as_pending(self, inventory_item: InventoryItem):
        pass


    def get_start_price(self, inventory_item: InventoryItem):
        return inventory_item.askStockX.start_price


    
    def is_ask_expired_on_marketplace(self, inventory_item: InventoryItem) -> bool:
        return False

    
    def refresh_expiry_date(self, inventory_item: InventoryItem):
        pass        