from datetime import datetime
from uuid import uuid4

from suplexedlibrary.database.postgres.postgres import get_session
from suplexedlibrary.exceptions import FailedFetchingListings
from suplexedlibrary.pricing.ConsignmentPricer import ConsignmentPricer
from suplexedlibrary.api.StadiumGoodsApi import StadiumGoodsApi
from suplexedlibrary.model.AskStadiumGoods import AskStadiumGoods
from suplexedlibrary.model.StadiumGoodsAccount import StadiumGoodsAccount
from suplexedlibrary.model.InventoryItem import InventoryItem
from suplexedlibrary.model.DataSyncAccount import DataSyncAccount
from suplexedlibrary.model.ResellPricing import ResellPricing
from suplexedlibrary.constants.ListingStatus import ListingStatus
from suplexedlibrary.constants.UserEventType import UserEventType

class StadiumGoodsPricer(ConsignmentPricer):

    def __init__(self, account: DataSyncAccount, stadium_goods_account: StadiumGoodsApi):
        self.asks = {}
        self.listings = []
        self.created_listings = set()
        super().__init__(account)
        self.api = StadiumGoodsApi(stadium_goods_email=stadium_goods_account.email, stadium_goods_password=stadium_goods_account.password)
        self.is_listing_data_healthy = True
        self.hydrate_ask_map()


    def does_ask_need_update(self, inventory_item: InventoryItem) -> bool:
        print('checking matched ask')
        matched_ask = self.asks[inventory_item.askStadiumGoods.exchange_ask_id] if inventory_item.askStadiumGoods.exchange_ask_id in self.asks else None
        print('passed check')
        if not matched_ask:
            raise Exception(f'No ask found for {inventory_item.askStadiumGoods.exchange_ask_id}')
        
        return matched_ask.listingPrice != inventory_item.askStadiumGoods.ask_amount

    def is_at_bottom_market(self, shoe: InventoryItem, lowest_ask: int) -> bool:
        current_ask = shoe.askStadiumGoods.ask_amount
        exchange_ask = self.asks[shoe.askStadiumGoods.exchange_ask_id]['amount']
        if lowest_ask >= exchange_ask:
            print('\tLowest ask = ' + str(lowest_ask))
            print('\tCurrent ask = ' + str(current_ask))
            return True

        return False
    
    def is_at_floor(self, shoe: InventoryItem) -> bool:
        ask: AskStadiumGoods = shoe.askStadiumGoods
        if ask == None:
            raise Exception('This ask should not be None')

        if self.asks[ask.exchange_ask_id]['amount'] - 1 <= ask.ask_amount:
            print('ask of ' + str(ask.ask_amount) +
                  ' is already at the floor of ' + str(self.asks[ask.exchange_ask_id]['amount']))
            return True

        return False

    def is_ask_corrupt(self, shoe: InventoryItem):
        if shoe.askStadiumGoods == None:
            return True
        return False

    def is_bottom_market_our_shoe(self, inventory_item: InventoryItem, bottom_market: int):
        all_asks = get_session().query(AskStadiumGoods).filter(
            AskStadiumGoods.stadium_goods_href == inventory_item.askStadiumGoods.stadium_goods_href,
            AskStadiumGoods.account_id == inventory_item.account_id
        ).all()

        matched_asks = list(filter(lambda ask_key : (self.asks[ask_key].sku == inventory_item.shoeSku and self.asks[ask_key].size == inventory_item.shoeSize), self.asks))
        
        for ask_id in matched_asks:
            if self.asks[ask_id].listingPrice == bottom_market and self.asks[ask_id].serial == inventory_item.askStadiumGoods.exchange_ask_id:
                return True

        return False


    def update_inventory_item_ask(self, inventory_item: InventoryItem, new_ask: int):
        self.api.update_ask(
            ask=inventory_item.askStadiumGoods,
            amount=new_ask,
            sell_floor=inventory_item.askStadiumGoods.ask_amount
        )
        event_type = None

        if is_floor_update:
            event_type = UserEventType.SELL_FLOOR_PRICE_UPDATE
        else:
            event_type = UserEventType.STATIC_PRICE_UPDATE

        event = UserEvent(
            id                                 = str(uuid4()),
            account_id                         = self.account.id,
            inventory_id                       = inventory_item.id,
            event_type                         = event_type,
            event_title                        = f'Updating ask for {inventory_item.shoeName} size {inventory_item.shoeSize} on StadiumGoods',
            event_text                         = f'Your listing price has been automatically updated from ${inventory_item.askStadiumGoods.real_ask_amount:,}.00 to ${new_ask:,}.00 on Stadium Goods.',
            event_link                         = None,
            event_time                         = datetime.now(),
            event_image                        = inventory_item.image_url
        )
        get_session().add(event)
        get_session().commit()

    
    def get_lowest_ask(self, resell_pricing: ResellPricing):
        return resell_pricing.stadium_goods_ask

    def hydrate_ask_map(self):
        listings = []
        
        try:
            listings = self.api.get_all_listings()
        except FailedFetchingListings as e:
            print(f'caught listing fetch failure: {e}')
            self.is_listing_data_healthy = False
        except Exception as e:
            print(e)
            self.is_listing_data_healthy = False

        self.listings = listings
        for ask in listings:
            self.asks[ask.serial] = ask

    def get_ask_amount(self, inventory_item: InventoryItem) -> int:
        return inventory_item.askStadiumGoods.ask_amount

    def get_inventory(self):
        inventory_items = get_session().query(InventoryItem).filter(
            InventoryItem.account_id == self.account.id,
            InventoryItem.askStadiumGoods != None,
            InventoryItem.saleOrder == None
        ).all()

        
        def filter_valid_stadiumgoods_inventory_items(inventory_item: InventoryItem) -> bool:
            return inventory_item.askStadiumGoods and inventory_item.askStadiumGoods.ask_amount != None
        
        inventory_items = list(filter(lambda x : filter_valid_stadiumgoods_inventory_items(x), inventory_items))


        return inventory_items

    def is_lowest_ask_calculated(self, resell_pricing: ResellPricing):
        return resell_pricing.is_stadiumgoods_ask_calculated


    def has_active_listing_been_removed_from_the_marketplace(self, inventory_item: InventoryItem) -> bool:
        ask: AskStadiumGoods = inventory_item.askStadiumGoods
        if ask == None:
            raise Exception('This ask should not be None')

        if ask.status == ListingStatus.ACTIVE and ask.exchange_ask_id not in self.asks:
            print('The user must have deleted this on the stadiumgoods side, lets remove it from our db for sanity')
            get_session().delete(ask)
            get_session().commit()
            return True

        print('\tthis listing has not been removed from the marketplace')
        return False 

    def is_ready_for_reprice(self, inventory_item: InventoryItem) -> bool:
        if not inventory_item.askStadiumGoods.last_priced_time:
            print(f'\t{inventory_item.shoeName}, {inventory_item.shoeSize}, {inventory_item.askStadiumGoods.exchange_ask_id} has not been priced by us yet')
            return True

        hours_in_a_day = 24 * 60
        difference = datetime.now() - inventory_item.askStadiumGoods.last_priced_time
        divmod(difference.days * hours_in_a_day + difference.seconds, 60)

    def is_resell_price_valid(self, inventory_item: InventoryItem) -> bool:
        resell_pricing: ResellPricing = inventory_item.resellPricing

        if not resell_pricing.stadium_goods_web_ask or not resell_pricing.stadium_goods_ask:
            print('\tnull ask found')

        elif resell_pricing.stadium_goods_ask < resell_pricing.stadium_goods_web_ask:
            print(f'\twe have an ask mismatch:\n\t\tseller_portal_ask:{resell_pricing.stadium_goods_ask}\n\t\twebsite_ask:{resell_pricing.stadium_goods_web_ask}')
        else:
            print('\tportal ask seems to be right')

if __name__ == '__main__':
    account = get_session().query(DataSyncAccount).filter(
        DataSyncAccount.email == 'dan').first()
    pricer = StockxPricer(account, account.stockxAccount[0])