from typing import List
import traceback
from uuid import uuid4
from requests import session
from datetime import datetime

from suplexedlibrary.model.UserEvent import UserEvent
from suplexedlibrary.constants.UserEventType import UserEventType
from suplexedlibrary.pricing.Pricer import Pricer
from suplexedlibrary.model.DataSyncAccount import DataSyncAccount
from suplexedlibrary.model.ResellPricing import ResellPricing
from suplexedlibrary.model.InventoryItem import InventoryItem
from suplexedlibrary.model.StockxAccount import StockxAccount
from suplexedlibrary.model.AskStockX import AskStockX
from suplexedlibrary.api.StockxApi import StockXApi
from suplexedlibrary.database.postgres.postgres import get_session
from suplexedlibrary.inventory.InventoryManagerStockX import InventoryManagerStockX
from suplexedlibrary.constants.ListingStatus import ListingStatus
from suplexedlibrary.exceptions import FailedFetchingListings, CreateListingException, UpdatePriceException

def filter_inventory(inventory_item: InventoryItem):
    if inventory_item.pricingStrategy.mute_stockx:
        return False
    return True


class StockxPricer(Pricer):
    ASK_ID = 'stockxAskId'

    def __init__(self, account: DataSyncAccount, stockx_account: StockxAccount):
        self.asks = {}
        self.recent_sales_asks = {}
        self.recent_sales = []
        self.listings = []
        self.created_listings = set()
        self.db_session = get_session()
        super().__init__(account)
        self.api = StockXApi(stockx_account, skip_login=False)
        self.is_listing_data_healthy = True
        self.hydrate_ask_map()
        self.hydrate_recent_sales_map()


    def process_missing_asks(self):
        if self.is_listing_data_healthy:
            InventoryManagerStockX.process_missing_listings_static(self.listings, self.recent_sales, self.account.id, self.api, self.created_listings)

    def is_ask_inactive(self, inventory_item: InventoryItem):
        return inventory_item.askStockX.status == ListingStatus.INACTIVE

    def is_ask_deleted(self, inventory_item: InventoryItem):
        return inventory_item.askStockX.status == ListingStatus.DELETE

    def does_ask_need_update(self, inventory_item: InventoryItem) -> bool:
        matched_ask = self.asks[inventory_item.askStockX.exchange_ask_id] if inventory_item.askStockX.exchange_ask_id in self.asks else None
        matched_ask_amount = matched_ask['amount']
        if not matched_ask:
            raise Exception(f'No ask found for {inventory_item.askStockX.exchange_ask_id}')
        
        print(f'{matched_ask_amount} == {inventory_item.askStockX.ask_amount} is {matched_ask_amount == inventory_item.askStockX.ask_amount}')

        if self.account.email == 'daniel.revie1@gmail.com':
            return float(matched_ask_amount) != float(inventory_item.askStockX.ask_amount) and inventory_item.askStockX.askAmount < matched_ask_amount
        
        return float(matched_ask_amount) != float(inventory_item.askStockX.ask_amount)

    def is_at_bottom_market(self, shoe: InventoryItem, lowest_ask: int) -> bool:
        current_ask = shoe.askStockX.ask_amount
        exchange_ask = self.asks[shoe.askStockX.exchange_ask_id]['amount']
        if lowest_ask >= exchange_ask:
            print('\tLowest ask = ' + str(lowest_ask))
            print('\tCurrent ask = ' + str(current_ask))
            return True

        return False

    def is_listed_on_exchange(self, shoe: InventoryItem) -> bool:
        if not shoe.askStockX:
            return False
        return shoe.askStockX.exchange_ask_id in self.asks

    def is_at_floor(self, shoe: InventoryItem) -> bool:
        ask: AskStockX = shoe.askStockX
        if ask == None:
            raise Exception('This ask should not be None')

        if self.asks[ask.exchange_ask_id]['amount'] - 1 < ask.ask_amount:
            print('ask of ' + str(ask.ask_amount) +
                  ' is already at the floor of ' + str(self.asks[ask.exchange_ask_id]['amount']))
            return True

        return False

    def is_ask_corrupt(self, shoe: InventoryItem):
        if shoe.askStockX == None:
            return True
        return False

    def is_bottom_market_our_shoe(self, inventory_item: InventoryItem, bottom_market: int):
        all_asks = self.db_session.query(AskStockX).filter(
            AskStockX.stx_sku_uuid == inventory_item.askStockX.stx_sku_uuid,
            AskStockX.account_id == inventory_item.account_id
        ).all()

        matched_asks = list(filter(lambda ask_key : (self.asks[ask_key]['product']['styleId'] == inventory_item.shoeSku and self.asks[ask_key]['product']['styleId']) or (not self.asks[ask_key]['product']['styleId'] and self.asks[ask_key]['product']['title'] == inventory_item.shoeName), self.asks))
        
        for ask_id in matched_asks:
            if self.asks[ask_id]['amount'] == bottom_market and self.asks[ask_id]['chainId'] == inventory_item.askStockX.exchange_ask_id:
                return True

        return False

    def delete_inventory_item_ask(self, inventory_item: InventoryItem):
        print(f'deleting exchange ask and ask object {inventory_item.id}')
        self.api.remove_ask(inventory_item.askStockX.exchange_ask_id)
        
        self.db_session.query(AskStockX).filter(AskStockX.id == inventory_item.askStockX.id).delete()
        # self.db_session.delete(inventory_item.askStockX)
        self.db_session.commit()

    def delete_exchange_ask(self, inventory_item: InventoryItem):
        try:
            print(f'deleting exchange ask {inventory_item.id}')
            self.api.remove_ask(inventory_item.askStockX.exchange_ask_id)
            inventory_item.askStockX.exchange_ask_id = None
            self.db_session.commit()
            print('\tdelete success')
        except Exception as e:
            print(e)
        

    def update_inventory_item_ask(self, inventory_item: InventoryItem, new_ask: int, is_floor_update: bool):
        self.api.update_ask(
            ask=inventory_item.askStockX,
            amount=new_ask,
            sell_floor=inventory_item.askStockX.ask_amount
        )
        event_type = None

        if is_floor_update:
            event_type = UserEventType.SELL_FLOOR_PRICE_UPDATE
        else:
            event_type = UserEventType.STATIC_PRICE_UPDATE

        event = UserEvent(
            id                                 = str(uuid4()),
            account_id                         = self.account.id,
            inventory_id                       = inventory_item.id,       
            sale_order_id                      = None,
            purchase_order_id                  = None,
            event_type                         = event_type,
            event_title                        = f'Updating ask for {inventory_item.shoeName} size {inventory_item.shoeSize} on StockX',
            event_text                         = f'Your listing price has been automatically updated from ${inventory_item.askStockX.real_ask_amount:,}.00 to ${new_ask:,}.00 on StockX.',
            event_link                         = None,
            event_time                         = datetime.now(),
            event_image                        = inventory_item.image_url
        )

        inventory_item.askStockX.real_ask_amount = new_ask
        self.db_session.add(event)
        self.db_session.commit()

    def list_inventory_item_ask(self, inventory_item: InventoryItem, amount: int = None, start_price: int = None):
        

        if not inventory_item.shoeSku:
            stx_url = StockXApi.get_stockx_url_from_algolia_search(inventory_item.shoeName)
        else:
            stx_url = StockXApi.get_stockx_url_from_algolia_search(inventory_item.shoeSku)


        ask_amount = inventory_item.askStockX.ask_amount

        if amount: # Amount override for floor pricing
            ask_amount = amount

        if start_price:
            ask_amount = start_price

        try:
            (exchange_ask_id, stx_sku_uuid) = self.api.place_new_ask(
                style_id=inventory_item.shoeSku,
                stx_url=stx_url,
                size=inventory_item.shoeSize,
                new_ask=ask_amount,
                sell_floor=ask_amount
            )
        except CreateListingException as e:
            print(e)
            self.db_session.add(UserEvent(
                id                                 = str(uuid4()),
                account_id                         = self.account.id,
                inventory_id                       = inventory_item.id,       
                sale_order_id                      = None,
                purchase_order_id                  = None,
                event_type                         = UserEventType.ERROR,
                event_title                        = f'Failed creating new ask for {inventory_item.shoeName} size {inventory_item.shoeSize} on StockX',
                event_text                         = f'Failed creating a new ask for ${ask_amount:,}.00 on StockX.',
                event_link                         = None,
                event_time                         = datetime.now(),
                event_image                        = inventory_item.image_url            
            ))
            return

        self.db_session.add(UserEvent(
            id                                 = str(uuid4()),
            account_id                         = self.account.id,
            inventory_id                       = inventory_item.id,       
            sale_order_id                      = None,
            purchase_order_id                  = None,
            event_type                         = UserEventType.CREATED_NEW_ASK,
            event_title                        = f'Creating new ask for {inventory_item.shoeName} size {inventory_item.shoeSize} on StockX',
            event_text                         = f'You have posted a new listing for sale on StockX.',
            event_link                         = None,
            event_time                         = datetime.now(),
            event_image                        = inventory_item.image_url            
        ))

        inventory_item.askStockX.status = ListingStatus.ACTIVE
        self.created_listings.add(exchange_ask_id)

        if inventory_item.askStockX:
            inventory_item.askStockX.exchange_ask_id = exchange_ask_id
            inventory_item.askStockX.real_ask_amount = ask_amount
        else:
            ask = AskStockX(
                id=str(uuid4()),
                inventory_item_id=inventory_item.id,
                ask_amount=ask_amount,
                real_ask_amoun=ask_amount,
                exchange='stockx',
                exchange_ask_id=exchange_ask_id,
                stx_sku_uuid=stx_sku_uuid,
                size=float(inventory_item.shoeSize),
                style_id=inventory_item.shoeSku,
                status=ListingStatus.ACTIVE,
            )
            self.db_session.add(ask)

        self.db_session.commit()

    def remove_all_asks(self):
        inventory_items = list(
            filter(lambda item: item.askStockX, self.account.inventory))
        for item in inventory_items:
            self.api.remove_ask(item.askStockX.exchange_ask_id)
            self.db_session.query(AskStockX).filter(AskStockX.id == item.askStockX.id).delete()
            self.db_session.commit()

    def get_lowest_ask(self, resell_pricing: ResellPricing):
        return resell_pricing.stockx_ask

    def hydrate_ask_map(self):
        listings = []

        try:
            listings = self.api.get_all_sales(sale_type='current')
        except FailedFetchingListings as e:
            print(f'caught listing fetch failure: {e}')
            print(f'\tfailed fetching current REST sales')
            self.is_listing_data_healthy = False
            raise e
        except Exception as e:
            print(e)
            print(f'uncaught listing fetch failure: {e}')
            print(f'\tfailed fetching current REST sales')            
            self.is_listing_data_healthy = False
            raise e

        self.listings = listings
        for ask in listings:
            self.asks[ask['chainId']] = ask

    def hydrate_recent_sales_map(self):
        recent_sales = []
        
        try:
            recent_sales = self.api.get_sales_history_gql()
        except FailedFetchingListings as e:
            print(f'caught listing fetch failure: {e}')
            print(f'\tfailed fetching historical GQL sales')
            self.is_listing_data_healthy = False
            raise e
        except Exception as e:
            print(f'uncaught listing fetch failure: {e}')
            print(f'\tfailed fetching historical GQL sales')            
            print(e)
            self.is_listing_data_healthy = False
            raise e

        self.recent_sales = recent_sales
        for recent_sale in recent_sales:
            self.recent_sales_asks[recent_sale['chainId']] = recent_sales

    def get_ask_amount(self, inventory_item: InventoryItem) -> int:
        return inventory_item.askStockX.ask_amount

    def get_inventory(self):
        inventory_items = self.db_session.query(InventoryItem).filter(
            InventoryItem.account_id == self.account.id,
            InventoryItem.askStockX != None
        ).all()
        count = self.db_session.query(InventoryItem).filter(
            InventoryItem.account_id == self.account.id,
            InventoryItem.askStockX != None,
        ).count()
        print(f'found {count} items')

        '''
        def filter_valid_goat_inventory_items(inventory_item: InventoryItem) -> bool:
            return inventory_item.askStockX and inventory_item.askStockX.ask_amount != None
        
        inventory_items = list(filter(lambda x : filter_valid_goat_inventory_items(x), inventory_items))
        '''
        return inventory_items

    def is_lowest_ask_calculated(self, resell_pricing: ResellPricing):
        return resell_pricing.is_stockx_ask_calculated

    def has_active_listing_been_removed_from_the_marketplace(self, inventory_item: InventoryItem) -> bool:
        ask: AskStockX = inventory_item.askStockX
        if ask == None:
            raise Exception('This ask should not be None')

        if ask.status == ListingStatus.ACTIVE and ask.exchange_ask_id not in self.asks and ask.exchange_ask_id not in self.recent_sales_asks:
            print('The user must have deleted this on the stockx side, lets remove it from our db for sanity')

            user_event = UserEvent(
                id                                 = str(uuid4()),
                account_id                         = self.account.id,
                inventory_id                       = inventory_item.id,       
                sale_order_id                      = None,
                purchase_order_id                  = None,               
                event_type                         = UserEventType.LISTING_REMOVE_OFF_ASSIST,
                event_title                        = f'Active listing removed off of sync platform for {inventory_item.shoeName} size {inventory_item.shoeSize} on StockX',
                event_text                         = f'Active listing removed off of sync platform for {inventory_item.shoeName} size {inventory_item.shoeSize} on StockX',
                event_link                         = None,
                event_time                         = datetime.now(),
                event_image                        = inventory_item.image_url                
            )
            self.db_session.add(user_event)         
            
            self.db_session.delete(ask)
            inventory_item.synced = False
            self.db_session.commit()
            return True

        return False 


    def is_listing_expired(self, inventory_item: InventoryItem):
        if inventory_item.askStockX.exchange_ask_id not in self.asks:
            return False

        stockx_ask = self.asks[inventory_item.askStockX.exchange_ask_id]
        now = datetime.now()
        expiry_date = datetime.strptime(stockx_ask['expiresAt'].split('+')[0], '%Y-%m-%dT%H:%M:%S')

        if expiry_date < now:
            print('this ask has expired')
            return True

        return False

    def mark_listing_as_pending(self, inventory_item: InventoryItem):
        stockx_ask: AskStockX = inventory_item.askStockX
        stockx_ask.status = ListingStatus.PENDING
        self.db_session.commit()


    def get_start_price(self, inventory_item: InventoryItem):
        return inventory_item.askStockX.start_price

    def is_ask_expired_on_marketplace(self, inventory_item: InventoryItem) -> bool:
        return inventory_item.askStockX.is_expired == True

    
    def refresh_expiry_date(self, inventory_item: InventoryItem):
        stockx_ask: AskStockX = inventory_item.askStockX
        order_details = self.api.get_sell_order_details(stockx_ask.exchange_ask_id)
        try:
            self.api.update_ask(
                ask=inventory_item.askStockX,
                amount=stockx_ask.real_ask_amount,
                sell_floor=inventory_item.askStockX.ask_amount,
                existing_item=order_details
            )
        except Exception as e:
            traceback.print_exc()
