import re
import random
import time
import operator
from uuid import uuid4
from typing import List
from datetime import datetime, timedelta
from sqlalchemy import or_
from concurrent.futures import ThreadPoolExecutor

from suplexedlibrary.model.ReportCardResult import ReportCardResult, ReportCardType
from suplexedlibrary.model.DataSyncAccount import DataSyncAccount
from suplexedlibrary.model.SaleOrder import SaleOrder
from suplexedlibrary.model.PurchaseOrder import PurchaseOrder
from suplexedlibrary.model.AskGoat import AskGoat
from suplexedlibrary.model.AskStockX import AskStockX
from suplexedlibrary.model.AskStadiumGoods import AskStadiumGoods
from suplexedlibrary.model.AskFlightClub import AskFlightClub
from suplexedlibrary.model.GoatSale import GoatSale
from suplexedlibrary.database.postgres.postgres import get_session
from suplexedlibrary.api.StockxApi import StockXApi
from suplexedlibrary.api.SuplexApi import SuplexApi
from suplexedlibrary.api.GoatApi import GoatApi
from suplexedlibrary.database.mongo.nosql import get_sneaker_db
from suplexedlibrary.exceptions import Scraping403Exception

suplexed_api = SuplexApi()



def filter_sales(sale: GoatSale, sold_date: str):
    # print('\t' + str(sale.purchased_at.date()))
    # print('\t\t' + str(sold_date - timedelta(days=21)))
    # print('\t\t' + str(sold_date + timedelta(days=21)))

    return sale.purchased_at.date() > sold_date - timedelta(days=21) or sale.purchased_at.date() < sold_date + timedelta(days=21)

def get_maximum_goat_sale(sales: List[GoatSale]) -> GoatSale:
    sorted_sales = sorted(sales, key=operator.attrgetter('price_usd_cents'))
    return sorted_sales[0]

def get_maximal_sale_possibilty_stockx(
    sale_order: SaleOrder, 
    api: StockXApi , 
    start_date: datetime, 
    end_date: datetime,
) -> int:

    stockx_url = None


    # TODO: parse new api
    suplexed_data = suplexed_api.get_pricing_data_v2_types(sale_order.shoe_sku.replace(' ', '-'))

    if 'stx_url' not in suplexed_data or not suplexed_data['stx_url']:
        stockx_url = StockXApi.get_stockx_url_from_algolia_search(sale_order.shoe_sku)
    else:
        stockx_url = suplexed_data['stx_url']


    child_id = get_sneaker_db().stockxSizes.find_one({'style_id': sale_order.shoe_sku, 'size': sale_order.shoe_size})
    if not child_id:
        attempt_count = 0

        while attempt_count < 10:
            try:
                child_id = StockXApi.get_stockx_sku_uuid_static(stockx_url, '{0:g}'.format(float(sale_order.shoe_size)))
                break
            except Scraping403Exception as e:
                print(e)
                attempt_count += 1
                time.sleep(random.randint(0, 3))
            except Exception as e:
                print(f'fatal exception: {e}')
                break


    sales_in_window = StockXApi.get_stockx_chart_sales_parsed(
        child_id, 
        sale_order.shoe_size, 
        sale_order.shoe_size, 
        start_date, 
        end_date
    )
    sales_in_window = list(filter(lambda x : x['average_sale_price'], sales_in_window))
    max_sale = None

    if sales_in_window:
        return max(sales_in_window, key=lambda x : x['average_sale_price'])['average_sale_price']

    return 0


def get_maximal_sale_possibilty_goat(
    sale_order: SaleOrder,
    start_date: datetime, 
    end_date: datetime,
) -> int:
    db = get_sneaker_db()
    tokens = list(map( lambda x : x['authorization'], db.goatLogins.find({})))
    sold_date = sale_order.sold_date
    goat_slug = db.goatAlgoliaProducts.find_one({'sku': sale_order.shoe_sku.replace('-', ' ')})['slug']
    if not goat_slug:
        print('no goat slug')
        return -1

    attempt_count = 0
    historical_goat_sales = []
    while attempt_count < 10:
        try:
            historical_goat_sales = GoatApi.get_goat_historical_sales(goat_slug, sale_order.shoe_size, random.choice(tokens))
            break
        except Scraping403Exception as e:
            print(e)
            attempt_count += 1
            time.sleep(random.randint(0, 4))
        except Exception as e:
            print(e)

    historical_goat_sales = list(filter(
        lambda x: datetime.strptime(x['start_date'], '%Y-%m-%dT%H:%M:%SZ').date() >= start_date and end_date <= (datetime.strptime(x['start_date'], '%Y-%m-%dT%H:%M:%SZ') + timedelta(days=7)).date(), historical_goat_sales['weekly_all_time']))
    historical_goat_sales = list(map(lambda x : float(x['average_price_cents']) / 100, historical_goat_sales))
    max_sale = max(historical_goat_sales)
    return max_sale

def get_maximal_sale_possibilty_stadiumgoods(sale_order: SaleOrder) -> int:
    sold_date = sale_order.sold_date

    return 0

def get_maximal_sale_possibilty_flightclub(sale_order: SaleOrder) -> int:
    sold_date = sale_order.sold_date

    return 0

def process_report_card_sale_result(sync_account: DataSyncAccount, sale_order: SaleOrder, report_card: ReportCardResult):
    try:
        # markets_to_measure_against = {'flightclub', 'stockx', 'goat', 'stadiumgoods'}
        sold_marketplace = sale_order.marketplace
        real_sale_amount = sale_order.sold_payout
        listing_start_date = sale_order.listing_creation_date.date() if sale_order.listing_creation_date else (sale_order.sold_date - timedelta(14))
        listing_sold_date = sale_order.sold_date
        sold_amount = sale_order.sold_amount

        if sold_marketplace == 'goat':
            real_sale_amount = real_sale_amount / 100

        
        stockx_api = StockXApi(sync_account.stockxAccount[0])
        try:
            maximal_profit_stockx = get_maximal_sale_possibilty_stockx(sale_order, stockx_api, listing_start_date, listing_sold_date)
        except Exception as e:
            print(f'failed fetching maximal profit for stockx')
            print(e)
        try:
            maximal_profit_goat = get_maximal_sale_possibilty_goat(sale_order, listing_start_date, listing_sold_date)
        except Exception as e:
            print(f'failed fetching maximal profit for goat')
            print(e)
        try:        
            maximal_profit_stadiumgoods = get_maximal_sale_possibilty_stadiumgoods(sale_order)
        except Exception as e:
            print(f'failed fetching maximal profit for stadiumgoods')
            print(e)
        try:        
            maximal_profit_flightclub = get_maximal_sale_possibilty_flightclub(sale_order)
        except Exception as e:
            print(f'failed fetching maximal profit for flightclub')
            print(e)

        maximal_profit = max(maximal_profit_stockx, maximal_profit_goat, maximal_profit_stadiumgoods, maximal_profit_flightclub)
        


        if maximal_profit > real_sale_amount:
            profit_forfeited = maximal_profit - real_sale_amount
            report_card.profit_missed_dollars += profit_forfeited
    except Exception as e:
        print(e)

